#############################################################################
# dvdshrink_plugin.tcl Copyright 2005, 2006, 2007, 2008 r0lZ
# PgcEdit plugin to help DVDShrink users.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "DVDShrink" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.0 or more running under Windows to use this plugin.
# Of course you need also DVDShrink 3.2 installed on your system to use this
# plugin.
#############################################################################
# Version History:
#
# v 1.2 (September 7, 2005)
# - First public release
# v 1.3 (November 25, 2005)
# - Added function to clear the Analysis cache of DVDShrink, to force it to
#   recompute the analysis of the already processed DVDs, or to preserve disc
#   space
# v 1.4 (January 22, 2006)
# - DVDShrink Remote Buttons: Added the options to use the Prev Chapter button
#   and/or the Return button to go back to the beginning of the previous title
# v 2.0 (June 2, 2006)
# - Totally new method used for the DVDShrink Remote Buttons function.  Now,
#   all VM commands are removed and recreated from scratch.
# - New option to go back to the last chapter of the previous title with the
#   prev chapter button.
# - The old method is kept unchanged, for users that prefer it.
# v 2.1 (September 15, 2006)
# - Bug in DVDShrink Remote Buttons's Always allow the 'Title Menu' button to
#   skip to the next title? function fixed.  The Title button was ineffective
#   in some cases.
# v 2.2 (April 19, 2007)
# - DVDShrink is now launched in the background when the "Launch DVDShrink"
#   menus are used.  The exit status of DVDShrink is no more checked.
# - Adapted to the new message dialogs of PgcEdit 8.1. Requires PgcEdit 8.1+!
# v 2.3 (May 25, 2007)
# - Removed the Cancel option in the dialog of Check Streams Remapping safety
#   as it was confusing and useless.
# v 2.4 (June 25, 2007)
# - Added the function "Fix compilation with 50 or more Titles" to fix the
#   wrong FP-PGC commands produced by Shrink when there are 50 or more Titles
#   in a compilation made in re-author mode.  As a bonus, the Title menu button
#   can be used to skip to the next Title.
# - Fixed a bug in "DVDShrink Remote Buttons" when there are more than 63 Titles
#   in the compilation and the GoUp button returns to the previous Title.
# v 2.5 (August 2, 2007)
# - Added "Enhance and fix authoring", a new method to assign navigation
#   functions to the remote buttons (including arrow buttons.)
# - Added "Make DVD completely region free", as DVDShrink still prohibits the
#   regions 7 (reserved) and 8 (special intl. venues).
# - Corrected some typos.
# v 2.6 (October 3, 2007)
# - Enhance and fix authoring uses now the original SetSTN commands defined by
#   the user with the DVD Shrink option "Set Default Streams..."
#############################################################################

if {! [winfo exists .mainmenu.plugins.dvdshrink]} {
	# Add Plugins -> DVDShrink submenu
	.mainmenu.plugins add cascade -label "DVDShrink" -menu .mainmenu.plugins.dvdshrink
	menu .mainmenu.plugins.dvdshrink
	.mainmenu.plugins.dvdshrink add command -label "Authoring erweitern und verbessern" -command {::plugins::dvdshrink::best_method}
	.mainmenu.plugins.dvdshrink add command -label "DVDShrink - Navigationstasten" -command {::plugins::dvdshrink::remote_buttons}
	.mainmenu.plugins.dvdshrink add command -label "Authoring nochmal!" -command {::plugins::dvdshrink::redo_authoring}
	.mainmenu.plugins.dvdshrink add command -label "Korrigiere Kompilation mit 50 oder mehr Titeln" -command {::plugins::dvdshrink::fix_50_titles}
	.mainmenu.plugins.dvdshrink add command -label "DVD komplett vom Regioncode befreien" -command {::plugins::dvdshrink::region_free}
	.mainmenu.plugins.dvdshrink add separator
	.mainmenu.plugins.dvdshrink add command -label "berprfe Streams, zur sicheren Neuanordnung" -command {::plugins::dvdshrink::check_streams_remapping}
	if {![catch {package require registry}]} {
		.mainmenu.plugins.dvdshrink add separator
		.mainmenu.plugins.dvdshrink add command -label "Starte DVDShrink (mit der aktuellen DVD)" -command {::plugins::dvdshrink::launch_dvdshrink true}
		.mainmenu.plugins.dvdshrink add command -label "Starte DVDShrink (leer)" -command {::plugins::dvdshrink::launch_dvdshrink false}
		.mainmenu.plugins.dvdshrink add separator
		.mainmenu.plugins.dvdshrink add command -label "Lsche DVDShrink-Cache: Ergebnisse der Analyse" -command {::plugins::dvdshrink::clear_analysis_cache}
	}
	.mainmenu.plugins.dvdshrink add separator
	.mainmenu.plugins.dvdshrink add command -label "ber..." -command {
			tk_messageBox  -title $::plugins::dvdshrink::plugintitle -icon info \
					-message "$::plugins::dvdshrink::plugintitle fr PgcEdit 8.1 oder hher.\nVersion $::plugins::dvdshrink::version\nAutor: r0lZ\nFreeware!"
		}
}

namespace eval ::plugins {}
namespace eval ::plugins::dvdshrink {

set version "2.6"
set plugintitle "DVDShrink Plugin"

#############################################################################
## add title button to skip to the next title on a DVD reauthored with DVDShrink
proc remote_buttons {} {
	if {! [test_version 8.1 $::plugins::dvdshrink::plugintitle]} return
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	if {$::pgcs(numvts) > 63} {
		set txt "Sorry, es gibt zu viele Titel in der Kompilation, um \"DVDShrink - Navigationstasten\" auszufhren."
		set txt "$txt\nBitte benutzen Sie stattdessen \"Authoring nochmal!\" (oder \"Korrigiere Kompilation mit 50 oder mehr Titeln\")."
		tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon info -message $txt \
			-parent .
		return
	}
	set txt "Navigationstasten fr Filme im DVDShrink-Modus \"Eigene DVD\"."
	set txt "$txt\n\nDieses Plugin arbeitet nur bei DVDs, die mit DVDShrink" 
	set txt "$txt\nim Modus \"Eigene DVD\" erstellt wurden!"
	set txt "$txt\n"
	set txt "$txt\nDieses Makro wird die Commands von (fast) jedem PGC"
	set txt "$txt\nmodifizieren und erstellt das VMGM Title Men PGC, so dass"
	set txt "$txt\nder Benutzer den \"Title\" Button der Fernbedienung benutzen"
	set txt "$txt\nkann, um zum nchsten Titel zu springen."
	set txt "$txt\n"
	set txt "$txt\nDank geht an 2COOL fr die Idee. Siehe diesen Thread:"
	set txt "$txt\nhttp://forum.doom9.org/showthread.php?threadid=61238"
	set txt "$txt\n"
	set txt "$txt\nDas Makro erlaubt es ebenso, die \"Prev\" und \"Next\" Chapter"
	set txt "$txt\nButtons der Fernbedienung zu benutzen, um zum ersten und"
	set txt "$txt\nletzten Kapitel eines Titel zu springen und den \"Return\" Button,"
	set txt "$txt\num zum Anfang des aktuellen Titel (unabhngig vom akt. Kapitel)"
	set txt "$txt\nzurckzugehen."
  set txt "$txt\nDie letzte Option erlaubt, automatisch eine Schleife (Loop)"
	set txt "$txt\nzum Anfang der DVD zu setzen, wenn der letzte Titel fertig"
	set txt "$txt\nabgespielt wurde."
	set txt "$txt\n"
	set txt "$txt\nWARNUNG: Nur einige grundlegende berprfungen werden durchgefhrt,"
	set txt "$txt\num sicherzustellen, dass die DVD im richtigen Format ist."
	set txt "$txt\nBenutzen Sie dieses Makro nur bei DVDs, die mit DVDShrink im"
	set txt "$txt\nModus \"Eigene DVD\" erstellt wurden!"
  set txt "$txt\nImmer ein Backup der IFOs erstellen, bevor Sie das Makro benutzen!"
	set txt "$txt\n"
	set txt "$txt\nFortfahren ?"
	set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon question -message $txt \
			-parent . -type yesno -default yes -skip dvdshrink_plugin_info]
		if {$r == "no"} return

	### init GUI defaults to reflect current state
	if {$::menus(0,numlus)==1 && $::menus(0,1,numpgcs)>=1 && $::menus(0,1,1,mask)==130} {
		set ::macros::dvdshrinkbuttons_title 1
	} else {
		set ::macros::dvdshrinkbuttons_title 0
	}

	set ::macros::dvdshrinkbuttons_prev [get_int ::pgcs(1,1,header) 0x9E 2]
	if {$::macros::dvdshrinkbuttons_prev > 2} {set ::macros::dvdshrinkbuttons_prev 1}

	set ::macros::dvdshrinkbuttons_next [get_int ::pgcs(1,1,header) 0x9C 2]
	if {$::macros::dvdshrinkbuttons_next > 1} {set ::macros::dvdshrinkbuttons_next 1}

	set ::macros::dvdshrinkbuttons_return [get_int ::pgcs(1,1,header) 0xA0 2]
	if {$::macros::dvdshrinkbuttons_return > 2} {set ::macros::dvdshrinkbuttons_return 1}

	if {[lsearch [get_type_cmds post ::pgcs($::pgcs(numvts),1)] \
			{30 08 00 00 01 00 00 00}] >= 0} {	;# CallSS FP
		set ::macros::dvdshrinkbuttons_loop 1
	} else {
		set ::macros::dvdshrinkbuttons_loop 0
	}

	### GUI
	if {[winfo exists .dvdshrinkbuttons]} {destroy .dvdshrinkbuttons}
	set w .dvdshrinkbuttons
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $::plugins::dvdshrink::plugintitle
	if {$::config(tk) != "aqua"} {wm transient $w .}
	wm protocol $w WM_DELETE_WINDOW {set ::macros::dvdshrinkbuttons_done 0}
	bind $w <Escape> "set ::macros::dvdshrinkbuttons_done 0"
	bind $w <Return> "set ::macros::dvdshrinkbuttons_done 1"

	set txt "Dieses Makro erlaubt, die Navigationstasten der Fernbedienung zum Navigieren in\nDVDs zu verwenden, die mit DVDShrink im Modus \"Eigene DVD\" erstellt wurden."
	label $w.lab -text $txt
	pack  $w.lab -padx 2 -pady 4

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -fill x
	checkbutton $w.f1.b2 -text "Wenn im 1.Kapitel, erlaube dem Button (Prev Chapter), zum Anfang des aktuellen Titel zurckzugehen?" \
			-variable ::macros::dvdshrinkbuttons_prev
	pack  $w.f1.b2 -anchor w
	checkbutton $w.f1.b2b -text "Wenn im 1.Kapitel, erlaube dem Button (Prev Chapter), zum Anfang des vorherigen Titel zurckzugehen?" \
			-variable ::macros::dvdshrinkbuttons_prev -onvalue 2
	pack  $w.f1.b2b -anchor w
	checkbutton $w.f1.b3 -text "Wenn im letzten Kapitel, erlaube dem Button (Next Chapter), zum nchsten Titel zu springen?" \
			-variable ::macros::dvdshrinkbuttons_next
	pack  $w.f1.b3 -anchor w
	checkbutton $w.f1.b6 -text "Immer dem (Return) Button erlauben, zum Anfang des aktuellen Titel zurckzugehen?" \
			-variable ::macros::dvdshrinkbuttons_return
	pack  $w.f1.b6 -anchor w
	checkbutton $w.f1.b6b -text "Immer dem (Return) Button erlauben, zum Anfang des vorherigen Titel zurckzugehen?" \
			-variable ::macros::dvdshrinkbuttons_return -onvalue 2
	pack  $w.f1.b6b -anchor w
	checkbutton $w.f1.b1 -text "Immer dem Title-Men Button erlauben, zum nchsten Titel zu springen?" \
			-variable ::macros::dvdshrinkbuttons_title
	pack  $w.f1.b1 -anchor w
	checkbutton $w.f1.b4 -text "Am Ende des letzten Titels, automatisch eine Schleife (Loop) zurck zum ersten Titel?" \
			-variable ::macros::dvdshrinkbuttons_loop
	pack  $w.f1.b4 -anchor w
	button $w.f1.b5 -text "Alle berprfen" -width 14 -command {
		set ::macros::dvdshrinkbuttons_title  1
		set ::macros::dvdshrinkbuttons_return 2
		set ::macros::dvdshrinkbuttons_prev   1
		set ::macros::dvdshrinkbuttons_next   1
		set ::macros::dvdshrinkbuttons_loop   1
	}
	pack  $w.f1.b5 -anchor w -padx 4 -pady 4


	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "Fortfahren" -width 10 -command {set ::macros::dvdshrinkbuttons_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Abbrechen" -width 10 -command {set ::macros::dvdshrinkbuttons_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	focus $w
	grab set $w

	tkwait variable ::macros::dvdshrinkbuttons_done
	grab release $w
	destroy $w
	if {!$::macros::dvdshrinkbuttons_done} return

	if {$::macros::dvdshrinkbuttons_title || $::macros::dvdshrinkbuttons_prev==2 \
			|| $::macros::dvdshrinkbuttons_return==2} {
		### create or replace the VMGM menu
		set r "yes"
		if {$::menus(0,numlus)>0} {
			set txt "WARNUNG: Es gibt bereits ein VMGM Men!"
			set txt "$txt\nWenn Sie fortfahren, wird es durch ein neues ersetzt!"
			set txt "$txt\n"
			set txt "$txt\nTrotzdem fortfahren ?"
			set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon warning -message $txt \
					-parent . -type yesno -default yes -skip dvdshrink_plugin_replace_vmgm]
		}
		if {$r == "yes"} {
			set ::raw_sectors(0,VMGM_PGCI_UT) \
					[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]

			::readdvd::parse_PGCI_UT ::raw_sectors(0,VMGM_PGCI_UT) 0 ::menus
			::main::init_tree_menu
			if {$::macros::dvdshrinkbuttons_title && \
					($::macros::dvdshrinkbuttons_prev==2 || $::macros::dvdshrinkbuttons_return==2)} {
				::main::display_pgc menu 0 1 1
				::utils::new_pgc true false
				set vmgm_pgc 2
			} else {
				set vmgm_pgc 1
			}
			if {$::macros::dvdshrinkbuttons_prev==2 || $::macros::dvdshrinkbuttons_return==2} {
				set cmds {{* pre}}
				set line 1
				set maxvts $::pgcs(numvts)
				if {$maxvts > 63} {
					set maxvts 63
					::utils::new_pgc true false
					lappend cmds "20 D4 00 01 00 3F 00 0[expr $vmgm_pgc+1]"
					set cmds2 {{* pre}}
					for {set vts 64} {$vts < $::pgcs(numvts)} {incr vts} {
						lappend cmds2 "00 B1 00 01 00 [format %02X $vts] 00 [format %02X [expr ($vts*2)-125]]"
						lappend cmds2 "30 02 00 00 00 [format %02X $vts] 00 00"
					}
					lappend cmds2 "30 02 00 00 00 [format %02X $vts] 00 00"
					lappend cmds "00 00 00 00 00 00 00 00"
					#lappend cmds "30 02 00 00 00 [format %02X $vts] 00 00"
					set cmds2 [concat $cmds2 {{* post} {* cell}}]
					set ::menus(0,1,[expr $vmgm_pgc+1]) $cmds2
					set gotolineoffset 3
				} else {
					lappend cmds {00 00 00 00 00 00 00 00}
					set gotolineoffset 2
				}
				for {set vts 1} {$vts < $maxvts} {incr vts} {
					lappend cmds "00 B1 00 01 00 [format %02X $vts] 00 [format %02X [expr ($vts*2)+$gotolineoffset]]"
					lappend cmds "30 02 00 00 00 [format %02X $vts] 00 00"
				}
				lappend cmds "00 00 00 00 00 00 00 00"
				lappend cmds "30 02 00 00 00 [format %02X $vts] 00 00"
				set cmds [concat $cmds {{* post} {* cell}}]
				set ::menus(0,1,$vmgm_pgc) $cmds

				### for each VTS except last one, replace the pre command "Set gprm(0) = 0"
				### by "Set gprm(0) = <current VTS number>"
				for {set vts 1} {$vts < $::pgcs(numvts)} {incr vts} {
					set cmds [get_type_cmds pre ::pgcs($vts,1)]
					set idx [lsearch -regexp $cmds {71 00 00 00 .. .. 00 00}]
					if {$idx == -1} {
						set cmds [linsert $cmds 0 "71 00 00 00 00 [format %02X $vts] 00 00"]
					} else {
						lset cmds $idx "71 00 00 00 00 [format %02X $vts] 00 00"
					}
					set_type_cmds pre ::pgcs($vts,1) $cmds
				}
			}
		}
	} else {
		### delete the VMGM menu
		set r "yes"
		if {$::menus(0,numlus)>0} {
			set txt "WARNUNG: Das komplette VMGM Men wird gelscht!"
			set txt "$txt\n"
			set txt "$txt\nTrotzdem fortfahren ?"
			set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon warning -message $txt \
					-parent . -type yesno -default yes dvdshrink_plugin_delete_vmgm]
			if {$r == "no"} return
		}
		if {$r == "yes"} {
			set ::raw_sectors(0,VMGM_PGCI_UT) ""
			set ::menus(0,numlus) 0
			::main::init_tree_menu

			### for each VTS except last one, replace the pre command
			### "Set gprm(0) = <current VTS number>" by "Set gprm(0) = 0"
			for {set vts 1} {$vts < $::pgcs(numvts)} {incr vts} {
				set cmds [get_type_cmds pre ::pgcs($vts,1)]
				set idx [lsearch -regexp $cmds {71 00 00 00 .. .. 00 00}]
				if {$idx != -1} {
					lset cmds $idx "71 00 00 00 00 00 00 00"
				}
				set_type_cmds pre ::pgcs($vts,1) $cmds
			}
		}
	}

	if {$::macros::dvdshrinkbuttons_loop} {
		### Last VTS: replace Exit by CallSS FP
		set vts $::pgcs(numvts)
		set cmds [get_type_cmds post ::pgcs($vts,1)]
		if {$cmds == ""} {
			set cmds {{30 08 00 00 01 00 00 00}}
		} else {
			set idx [lsearch $cmds {30 01 00 00 00 00 00 00}]	;# Exit
			if {$idx >= 0} {
				lset cmds $idx "30 08 00 00 01 00 00 00"	;# CallSS FP
			}
		}
		set_type_cmds post ::pgcs($vts,1) $cmds
	} else {
		### Last VTS: replace CallSS-FP by Exit
		set vts $::pgcs(numvts)
		set cmds [get_type_cmds post ::pgcs($vts,1)]
		if {$cmds == ""} {
			set cmds {{30 01 00 00 00 00 00 00}}
		} else {
			set idx [lsearch $cmds {30 08 00 00 01 00 00 00}]	;# CallSS FP
			if {$idx >= 0} {
				lset cmds $idx "30 01 00 00 00 00 00 00"	;# Exit
			}
		}
		set_type_cmds post ::pgcs($vts,1) $cmds
	}

	### Allow to use the Prev button, when pressed during Chapter 1) to go back to the
	### beginning of the title, and the Next Button to go directly to the end of the Title
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set goback false
		if {$::macros::dvdshrinkbuttons_prev == 0} {
			set_int ::pgcs($vts,1,header) 0x9E 0 2
		} else {
			set prevpgc $::macros::dvdshrinkbuttons_prev
			set_int ::pgcs($vts,1,header) 0x9E $prevpgc 2
			if {$prevpgc == 2} {set goback true}
		}
		if {! $::macros::dvdshrinkbuttons_next} {
			set_int ::pgcs($vts,1,header) 0x9C 0 2
		} else {
			set_int ::pgcs($vts,1,header) 0x9C 1 2
		}
		if {$::macros::dvdshrinkbuttons_return == 0} {
			set_int ::pgcs($vts,1,header) 0xA0 0 2
		} else {
			set prevpgc $::macros::dvdshrinkbuttons_return
			set_int ::pgcs($vts,1,header) 0xA0 $prevpgc 2
			if {$prevpgc == 2} {set goback true}
		}
		::main::display_pgc pgc $vts 1
		if {$goback} {
			# create new dummy PGC
			if {$::pgcs($vts,numpgcs) == 1} {
				::utils::new_pgc true false
			}
			set vmgm_pgc [expr $::macros::dvdshrinkbuttons_title+1]
			set targetvts [expr $vts-1]
			if {$targetvts == 0} {set targetvts $::pgcs(numvts)}
			set ::pgcs($vts,2) [list "* pre" "71 00 00 01 00 [format %02X $targetvts] 00 00" "30 08 00 0$vmgm_pgc 01 C0 00 00" "* post" "* cell"]
		} else {
			if {$::pgcs($vts,numpgcs) == 2} {
				#set tmp [info exists ::warning_delete_last_pgc]
				#set ::warning_delete_last_pgc true
				::utils::delete_last_pgc true
				#if {!$tmp} {
				#	unset ::warning_delete_last_pgc
				#}
			}
		}
	}

	### util done.  Refresh listbox.
	refresh_current_pgc
	::main::dvd_modified true false

	# if there is an old VMGM VOB, delete it or move it to backup
	::utils::delete_vob_with_confirm "VIDEO_TS.VOB" true
}

proc redo_authoring {} {
	if {! [test_version 8.1 $::plugins::dvdshrink::plugintitle]} return
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set ::plugins::dvdshrink::redo_title 1
	set ::plugins::dvdshrink::redo_goup  4
	set ::plugins::dvdshrink::redo_prev  1
	set ::plugins::dvdshrink::redo_next  3
	set ::plugins::dvdshrink::redo_exit  0

	### GUI
	set w .dvdshrinkredo
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $::plugins::dvdshrink::plugintitle
	if {$::config(tk) != "aqua"} {wm transient $w .}
	wm protocol $w WM_DELETE_WINDOW {set ::plugins::dvdshrink::redo_done 0}
	bind $w <Escape> "set ::plugins::dvdshrink::redo_done 0"
	bind $w <Return> "set ::plugins::dvdshrink::redo_done 1"

	set txt "Nochmaliges Authoring einer DVD, erstellt durch DVDShrink im Modus \"Eigene DVD\"."
	label $w.lab -text $txt
	pack  $w.lab -padx 2 -pady 4

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -fill x
	label $w.f1.lab -text "Wenn der Title-Button gedrckt wird..." -pady 0
	pack  $w.f1.lab -anchor w
	radiobutton $w.f1.b1 -text "springe zum Anfang des nchsten Titel." \
			-variable ::plugins::dvdshrink::redo_title -value 1 -pady 0 -bd 0
	pack  $w.f1.b1 -anchor w
	radiobutton $w.f1.b0 -text "nichts tun." \
			-variable ::plugins::dvdshrink::redo_title -value 0 -pady 0 -bd 0
	pack  $w.f1.b0 -anchor w

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x
	label $w.f2.lab -text "Wenn der GoUp (Return) Button gedrckt wird..." -pady 0
	pack  $w.f2.lab -anchor w
	radiobutton $w.f2.b1 -text "springe zum Anfang des vorherigen Titel." \
			-variable ::plugins::dvdshrink::redo_goup -value 4 -pady 0 -bd 0
	pack  $w.f2.b1 -anchor w
	radiobutton $w.f2.b2 -text "springe zum letzten Kapitel des vorherigen Titel." \
			-variable ::plugins::dvdshrink::redo_goup -value 2 -pady 0 -bd 0
	pack  $w.f2.b2 -anchor w
	radiobutton $w.f2.b3 -text "springe zum Anfang des aktuellen Titel." \
			-variable ::plugins::dvdshrink::redo_goup -value 1 -pady 0 -bd 0
	pack  $w.f2.b3 -anchor w
	radiobutton $w.f2.b0 -text "nichts tun." \
			-variable ::plugins::dvdshrink::redo_goup -value 0 -pady 0 -bd 0
	pack  $w.f2.b0 -anchor w

	frame $w.f3 -bd 2 -relief groove
	pack  $w.f3 -padx 4 -pady 2 -fill x
	label $w.f3.lab -text "Wenn der Prev-Chapter Button gedrckt wird, whrend das 1.Kapitel abgespielt wird..." -pady 0
	pack  $w.f3.lab -anchor w
	radiobutton $w.f3.b1 -text "springe zum Anfang des vorherigen Titel." \
			-variable ::plugins::dvdshrink::redo_prev -value 4 -pady 0 -bd 0
	pack  $w.f3.b1 -anchor w
	radiobutton $w.f3.b2 -text "springe zum letzten Kapitel des vorherigen Titel." \
			-variable ::plugins::dvdshrink::redo_prev -value 2 -pady 0 -bd 0
	pack  $w.f3.b2 -anchor w
	radiobutton $w.f3.b3 -text "springe zum Anfang des aktuellen Titel." \
			-variable ::plugins::dvdshrink::redo_prev -value 1 -pady 0 -bd 0
	pack  $w.f3.b3 -anchor w
	radiobutton $w.f3.b0 -text "nichts tun." \
			-variable ::plugins::dvdshrink::redo_prev -value 0 -pady 0 -bd 0
	pack  $w.f3.b0 -anchor w

	frame $w.f4 -bd 2 -relief groove
	pack  $w.f4 -padx 4 -pady 2 -fill x
	label $w.f4.lab -text "Wenn der Next-Chapter Button gedrckt wird, whrend das letzte Kapitel abgespielt wird..." -pady 0
	pack  $w.f4.lab -anchor w
	radiobutton $w.f4.b1 -text "springe zum darauffolgenden Titel." \
			-variable ::plugins::dvdshrink::redo_next -value 3 -pady 0 -bd 0
	pack  $w.f4.b1 -anchor w
	radiobutton $w.f4.b0 -text "nichts tun." \
			-variable ::plugins::dvdshrink::redo_next -value 0 -pady 0 -bd 0
	pack  $w.f4.b0 -anchor w

	frame $w.f5 -bd 2 -relief groove
	pack  $w.f5 -padx 4 -pady 2 -fill x
	label $w.f5.lab -text "Wenn der letzte Titel fertig abgespielt wurde..." -pady 0
	pack  $w.f5.lab -anchor w
	radiobutton $w.f5.b1 -text "Schleife (Loop) zurck zum Anfang des ersten Titel." \
			-variable ::plugins::dvdshrink::redo_exit -value 0 -pady 0 -bd 0
	pack  $w.f5.b1 -anchor w
	radiobutton $w.f5.b0 -text "beende die Wiedergabe der DVD." \
			-variable ::plugins::dvdshrink::redo_exit -value 1 -pady 0 -bd 0
	pack  $w.f5.b0 -anchor w

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 2 -fill x
	button $w.fb.b5 -text "Voreinstellungen" -width 16 -command {
		set ::plugins::dvdshrink::redo_title 1
		set ::plugins::dvdshrink::redo_goup  4
		set ::plugins::dvdshrink::redo_prev  1
		set ::plugins::dvdshrink::redo_next  3
		set ::plugins::dvdshrink::redo_exit  0
	}
	pack   $w.fb.b5 -side left -padx 2
	button $w.fb.cancel -text "Abbrechen" -width 10 -command {set ::plugins::dvdshrink::redo_done 0}
	pack   $w.fb.cancel -side right -padx 2
	button $w.fb.ok -text "OK" -width 10 -command {set ::plugins::dvdshrink::redo_done 1}
	pack   $w.fb.ok -side right -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	focus $w
	grab set $w

	tkwait variable ::plugins::dvdshrink::redo_done
	grab release $w
	destroy $w
	if {!$::plugins::dvdshrink::redo_done} return



	# Number of Titles (= number of VTS) in DVDShrink compilation
	set maxvts $::pgcs(numvts)

	set r "yes"
	if {$::menus(0,numlus)>0} {
		set txt "WARNUNG: Es gibt bereits ein VMGM Men!"
		set txt "$txt\nWenn Sie fortfahren, wird es durch ein neues ersetzt!"
		set txt "$txt\n"
		set txt "$txt\nTrotzdem fortfahren ?"
		set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon warning -message $txt \
				-parent . -type yesno -default yes -skip dvdshrink_plugin_replace_vmgm]
	}
	if {$r == "no"} {
		return
	}

	### create or replace the VMGM menu

	# FP-PGC - Init GPRMs and jump to title 1
	# 1  (JumpTT) Jump to Title 1
	set ::pgcs(0,0) [list \
		{* pre} \
		{30 02 00 00 00 01 00 00} \
		{* post} \
		{* cell} ]

	# Create a VMGM menu with 4 dummy PGCs
	set ::raw_sectors(0,VMGM_PGCI_UT) \
			[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]

	::readdvd::parse_PGCI_UT ::raw_sectors(0,VMGM_PGCI_UT) 0 ::menus
	::main::init_tree_menu
	if {!$::plugins::dvdshrink::redo_title} {
		::utils::set_menu_type 0 1 1 0
	}
	::main::display_pgc menu 0 1 1
	::utils::new_pgc true false
	::utils::new_pgc true false
	::utils::new_pgc true false

	# VMGM PGC 1 - Used to jump to the next title
	# 1  Set gprm(1) =(mov) 0
	# 2  Set gprm(0) =(mov) sprm(4:Title number in volume)
	# 3  Set gprm(0) +=(add) 1
	# 4  if ( gprm(0) > <number of VTS> ) then { Set gprm(0) =(mov) 1 }
	# 5  LinkPGCN PGC 4
	set ::menus(0,1,1) [list \
		{* pre} \
		{71 00 00 01 00 00 00 00} \
		{61 00 00 00 00 84 00 00} \
		{73 00 00 00 00 01 00 00} \
		"71 D0 00 00 00 01 00 [format %02X $maxvts]" \
		{20 04 00 00 00 00 00 04} \
		{* post} \
		{* cell} ]

	# VMGM PGC 2 - Used to jump to the beginning of the previous title
	# 1  Set gprm(1) =(mov) 0
	# 2  Set gprm(0) =(mov) sprm(4:Title number in volume)
	# 3  Set gprm(0) -=(add) 1
	# 4  if ( gprm(0) < 1 ) then { Set gprm(0) =(mov) <number of VTS> }
	# 5  LinkPGCN PGC 4
	set ::menus(0,1,2) [list \
		{* pre} \
		{71 00 00 01 00 00 00 00} \
		{61 00 00 00 00 84 00 00} \
		{74 00 00 00 00 01 00 00} \
		"71 F0 00 00 00 [format %02X $maxvts] 00 01" \
		{20 04 00 00 00 00 00 04} \
		{* post} \
		{* cell} ]

	# VMGM PGC 3 - Used to jump to the last chapter of the previous title
	# 1  Set gprm(1) =(mov) 1
	# 2  Set gprm(0) =(mov) sprm(4:Title number in volume)
	# 3  Set gprm(0) -=(add) 1
	# 4  if ( gprm(0) < 1 ) then { Set gprm(0) =(mov) <number of VTS> }
	# 5  LinkPGCN PGC 4
	set ::menus(0,1,3) [list \
		{* pre} \
		{71 00 00 01 00 01 00 00} \
		{61 00 00 00 00 84 00 00} \
		{74 00 00 00 00 01 00 00} \
		"71 F0 00 00 00 [format %02X $maxvts] 00 01" \
		{20 04 00 00 00 00 00 04} \
		{* post} \
		{* cell} ]

	# VMGM PGC 4 - Used to jump to the title pointed to by GPRM 0.
	# Since there cannot be more than 128 commands, if there are
	# more than 63 titles in the compilation, we need also the
	# VMGM PGC 5 to jump to titles 64 to 99.
	set cmds {{* pre}}
	set line 1
	set lastvts $maxvts
	if {$maxvts > 63} {
		set lastvts 63
		::utils::new_pgc true false	;# create VMGM PGC 5
		# First line of PGC 4 is:
		# 1  if ( gprm(0) > 63 ) then { LinkPGCN PGC 5 }
		lappend cmds "20 D4 00 00 00 3F 00 05"
		set cmds2 {{* pre}}
		for {set vts 64} {$vts <= $maxvts} {incr vts} {
			# if ( gprm(0) != <title number> ) then { Goto line <2 lines later> }
			lappend cmds2 "00 B1 00 00 00 [format %02X $vts] 00 [format %02X [expr ($vts*2)-125]]"
			# (JumpTT) Jump to Title <title number>
			lappend cmds2 "30 02 00 00 00 [format %02X $vts] 00 00"
		}
		lappend cmds2 {30 06 00 00 00 00 00 00}	;# (JumpSS) Jump to First Play PGC
		set ::menus(0,1,5) [concat $cmds2 {{* post} {* cell}}]
	} else {
		# Less than 64 commands: first line of VMGM PGC 4 is NOP
		lappend cmds {00 00 00 00 00 00 00 00}
	}
	for {set vts 1} {$vts <= $lastvts} {incr vts} {
		# if ( gprm(0) != <title number> ) then { Goto line <2 lines later> }
		lappend cmds "00 B1 00 00 00 [format %02X $vts] 00 [format %02X [expr ($vts*2)+2]]"
		# (JumpTT) Jump to Title <title number>
		lappend cmds "30 02 00 00 00 [format %02X $vts] 00 00"
	}
	lappend cmds {30 06 00 00 00 00 00 00}	;# (JumpSS) Jump to First Play PGC
	set ::menus(0,1,4) [concat $cmds {{* post} {* cell}}]


	### for each VTS
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		# compute the PTT # to use.
		# If the last PTT is a dummy chatper (length < 1 sec.), use the previous one.
		# Also, try to avoid illegal chapter numbers (if any).
		set pttlist $::pgcs($vts,1,pttlist)
		set programmap $::pgcs($vts,1,programmap)
		set numprograms [string length $programmap]
		set cellpbtable $::pgcs($vts,1,cellpbtable)
		set ptt [llength $pttlist]
		while {$ptt > 1} {
			set ptt [llength $pttlist]
			set lastpttentry [lindex $pttlist end]
			set lastpttpg [lindex $lastpttentry 0]
			if {$numprograms < $lastpttpg} {
				set pttlist [lreplace $pttlist end end]
				continue
			}
			set lastpttcell [get_int programmap [expr $lastpttpg-1] 1]
			if {$::pgcs($vts,1,numcells) < $lastpttcell || $lastpttcell == 0} {
				set pttlist [lreplace $pttlist end end]
				continue
			}
			if {$lastpttcell == $::pgcs($vts,1,numcells)} {
				set idx [expr (($lastpttcell-1)*24)+4]
				set bcdpart [string range $cellpbtable $idx [expr $idx+2]]
				if {$bcdpart == "\000\000\000"} {
					set pttlist [lreplace $pttlist end end]
					continue
				}
			}
			break
		}

		# replace the commands in the main titles PGCs:
		# *** pre commands:
		# <copy the original SetSTN commands, if any>
		# ?  if ( gprm(1) == 0 ) then { Break }
		# ?  LinkPTT Chapter <last (real) chapter in title>
		# *** post commands:
		# 1  (CallSS) Call the VMGM PGC 1, resume cell 1
		# (or exit post command in last VTS)
		set setstncmds [lsearch -all -inline -regexp $::pgcs($vts,1) {51 .. .. .. .. .. .. ..}]
		set ::pgcs($vts,1) {}
		lappend ::pgcs($vts,1) {* pre}
		foreach c $setstncmds {
			lappend ::pgcs($vts,1) $c
		}
		lappend ::pgcs($vts,1) {00 A2 00 01 00 00 00 00}
		lappend ::pgcs($vts,1) "20 05 00 00 00 00 0[format %01X [expr $ptt/256]] [format %02X [expr $ptt & 255]]"
		lappend ::pgcs($vts,1) {* post}
		if {$::plugins::dvdshrink::redo_exit && $vts == $maxvts} {
			lappend ::pgcs($vts,1) {30 01 00 00 00 00 00 00}	;# exit
		} else {
			lappend ::pgcs($vts,1) {30 08 00 01 01 C0 00 00}	;# (CallSS) Call the VMGM PGC 1, resume cell 1
		}
		lappend ::pgcs($vts,1) {* cell}

		# Create 3 dummy PGCs for use with the Prev, Next and GoUp PGCN links
		::main::display_pgc pgc $vts 1
		if {$::pgcs($vts,numpgcs) < 2} {
			::utils::new_pgc true false
		}
		if {$::pgcs($vts,numpgcs) < 3} {
			::utils::new_pgc true false
		}
		if {$::pgcs($vts,numpgcs) < 4} {
			::utils::new_pgc true false
		}
		# Dummy PGC #2 - Used for the PrevPGCN link
		# 1  (CallSS) Call the VMGM PGC 3, resume cell 1
		set ::pgcs($vts,2) [list \
			{* pre} \
			{30 08 00 03 01 C0 00 00} \
			{* post} \
			{* cell} ]
		# Dummy PGC #3 - Used for the NextPGCN link
		# 1  (CallSS) Call the VMGM PGC 1, resume cell 1
		set ::pgcs($vts,3) [list \
			{* pre} \
			{30 08 00 01 01 C0 00 00} \
			{* post} \
			{* cell} ]
		# Dummy PGC #4 - Used for the GoUpPGCN link
		# 1  (CallSS) Call the VMGM PGC 2, resume cell 1
		set ::pgcs($vts,4) [list \
			{* pre} \
			{30 08 00 02 01 C0 00 00} \
			{* post} \
			{* cell} ]

		# change the Prev, Next and GoUp PGCN links in the main title PGC
		set_int ::pgcs($vts,1,header) 0x9E $::plugins::dvdshrink::redo_prev 2
		set_int ::pgcs($vts,1,header) 0x9C $::plugins::dvdshrink::redo_next 2
		set_int ::pgcs($vts,1,header) 0xA0 $::plugins::dvdshrink::redo_goup 2
	}

	### util done.  Refresh listbox.
	::main::display_pgc pgc 0 0
	::main::dvd_modified true false

	# if there is an old VMGM VOB, delete it or move it to backup
	::utils::delete_vob_with_confirm "VIDEO_TS.VOB" true
}

#############################################################################
## Safety check for DVDShrink's Logical remapping of enabled streams option.
## Stored in: HKEY_CURRENT_USER\Software\DVD Shrink\DVD Shrink 3.2\Preferences\LogicalStreamRemapping
proc check_streams_remapping {} {
	if {! [test_version 8.1 $::plugins::dvdshrink::plugintitle]} return
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set used false

	# Titles
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set ttn 1} {$ttn <= $::pgcs($vts,numpgcs)} {incr ttn} {
			set used [search_setstn_gprm ::pgcs($vts,$ttn)]
			if {$used} break
		}
		if {$used} break
	}

	# Menus
	if {!$used} {
		for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					set used [search_setstn_gprm ::menus($vts,$lu,$pgc)]
					if {$used} break
				}
				if {$used} break
			}
			if {$used} break
		}
	}

	# Menu buttons
	if {!$used} {
		for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
			set butcmds {}
			if {[info exists ::buttons::butsinfo(menu,$vts,errorcode)] && \
					$::buttons::butsinfo(menu,$vts,errorcode) == 0} {
				for {set vid 1} {$vid <= $::buttons::butsinfo(menu,$vts,numvobids)} {incr vid} {
					for {set cid 1} {$cid <= $::buttons::butsinfo(menu,$vts,$vid,numcellids)} {incr cid} {
						if {[info exists ::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts)] && \
									$::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts) > 0} {
							for {set group 1} {$group <= $::buttons::butsinfo(menu,$vts,$vid,$cid,numgroups)} {incr group} {
								for {set but 1} {$but <= $::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts)} {incr but} {
									lappend butcmds $::buttons::butsinfo(menu,$vts,$vid,$cid,$group,$but,command)
								}
							}
						}
					}
				}
			}
			set used [search_setstn_gprm butcmds]
			if {$used} break
		}
		unset butcmds
	}

	# BOVs
	if {!$used} {
		for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
			set butcmds {}
			if {[info exists ::buttons::butsinfo(pgc,$vts,errorcode)] && \
					$::buttons::butsinfo(pgc,$vts,errorcode) == 0} {
				for {set vid 1} {$vid <= $::buttons::butsinfo(pgc,$vts,numvobids)} {incr vid} {
					for {set cid 1} {$cid <= $::buttons::butsinfo(pgc,$vts,$vid,numcellids)} {incr cid} {
						if {[info exists ::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)] && \
									$::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts) > 0} {
							for {set group 1} {$group <= $::buttons::butsinfo(pgc,$vts,$vid,$cid,numgroups)} {incr group} {
								for {set but 1} {$but <= $::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)} {incr but} {
									lappend butcmds $::buttons::butsinfo(pgc,$vts,$vid,$cid,$group,$but,command)
								}
							}
						}
					}
				}
			}
			set used [search_setstn_gprm butcmds]
			if {$used} break
		}
		unset butcmds
	}

	set txt "berprfe Streams, zur sicheren Neuanordnung\n\n"
	set txt "${txt}Dieses Makro berprft, ob Sie die DVDShrink-Option\n" 
  set txt "${txt}\"Logische Neuanordnung der ausgewhlten Spuren\"\n"
  set txt "${txt}mit der aktuellen DVD sicher benutzen knnen.\n\n"
  set txt "${txt}Wenn diese Option eingeschaltet ist, entfernt DVDShrink die\n"
	set txt "${txt}physiskalischen Definitionen der entfernten Streams in den IFOs\n"
	set txt "${txt}und versucht alle SetSTN-Commands zu ndern, um auf die neuen\n"
	set txt "${txt}Nummern der Streams zu verweisen.\n"
  set txt "${txt}Jedoch ist es nicht mglich, die indirekten SetSTN-Commands zu\n"
	set txt "${txt}ndern (wenn ein GPRM statt eines konstanten Wertes verwendet\n"
	set txt "${txt}wird), weil die Nummer des Stream im Command nicht fest\n"
	set txt "${txt}programmiert ist.\n"
  set txt "${txt}Es ist am besten, diese Option (wann immer es mglich ist)\n"
	set txt "${txt}einzuschalten oder es wird Referenzen zu den nicht vorhandenen\n" 
	set txt "${txt}Streams geben, die den DVD-Player verwirren knnen.\n"
  set txt "${txt}Das bedeutet, dass es besser ist, die Logische Neuanordnung im\n"
	set txt "${txt}Modus \"Komplette Disk\" zu benutzen, aber nur wenn es keine\n" 
	set txt "${txt}indirekte SetSTN-Commands in der DVD gibt.\n\n"
  set txt "${txt}Hinweise:\n"
	set txt "${txt}- Diese Option ist nutzlos, wenn Sie keinen Stream entfernen.\n"
	set txt "${txt}- Sie knnen es sicher benutzen, wenn Sie nur die letzten Streams entfernen.\n"
  set txt "${txt}- Die Option wird im DVDShrink-Modus \"Eigene DVD\" ignoriert.\n\n"
	set txt "${txt}Ergebnis der Analyse:\n\n"
	if {$used} {
		set txt "${txt}Es gibt indirekte SetSTN-Commands in dieser DVD.\n"
		set txt "${txt}(Wenn Sie die indirekten Commands finden wollen, suchen Sie\n"
		set txt "${txt}nach \"\[41\" in den Commands)\n"
    set txt "${txt}Die Option \"Logische Neuanordnung der ausgewhlten Spuren\"\n"
		set txt "${txt}sollte im Modus \"Komplette Disk\" AUSGESCHALTET werden!\n"
    show_msg "Option \"Logische Neuanordnung der ausgewhlten Spuren\"\nin DVDShrink nicht benutzen."
		set rec "AUS"
		set default "no"
	} else {
		set txt "${txt}Keine indirekten SetSTN-Commands in dieser DVD gefunden.\n"
		set txt "${txt}Option \"Logische Neuanordnung der ausgewhlten Spuren\"\n"
		set txt "${txt}kann sicher EINGESCHALTET werden.\n"
    show_msg "Option \"Logische Neuanordnung der ausgewhlten Spuren\"\nkann in DVDShrink benutzt werden."
		set rec "EIN"
		set default "yes"
	}
	set txt "${txt}\nZusammenfassung fr die aktuelle DVD:\n\n"
	set txt "${txt}Empfohlene Einstellungen fr den Modus \"Komplette Disk\":    $rec \n"

	set change false
	if {![catch {package require registry}]} {
		if {![catch {set cur [registry get \
				"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2 deutsch\\Preferences" \
				"LogicalStreamRemapping"]}]} {
			set change true
			if {$cur == 0} {
				set cur AUS
			} else {
				set cur EIN
			}
		}
	}
	if {$change} {
		set txt "${txt}Momentane Einstellung der Option in DVDShrink 3.2:    $cur\n\n"
		set txt "${txt}Option \"Logische Neuanordnung der ausgewhlten Spuren\"\n"
		set txt "${txt}in DVDShrink 3.2 erlauben ?"
		set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon info \
				-message $txt -parent . -type yesno -default $default]
		if {$r == "yes"} {
			catch {registry set \
				"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2 deutsch\\Preferences" \
				"LogicalStreamRemapping" 1}
		} elseif {$r == "no"} {
			catch {registry set \
				"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2 deutsch\\Preferences" \
				"LogicalStreamRemapping" 0}
		}
	} else {
		tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon info -message $txt -parent .
	}
	return $used
}
proc search_setstn_gprm {cmds_upvar} {
	upvar $cmds_upvar cmds
	if {[string first "\{41 " $cmds] == -1} {
		return false
	} else {
		return true
	}
}

#############################################################################
## Launch DVDShrink, with current DVD (if there is one DVD loaded).
proc launch_dvdshrink {with_current_dvd} {
	if {! [test_version 8.1 $::plugins::dvdshrink::plugintitle]} return
	if {![catch {package require registry}]} {
		if {![catch {set filename [registry get \
				"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2 deutsch\\Preferences" \
				"InstallApp"]}]} {
			if {[info exists ::dvddir] && $with_current_dvd} {
				if {$::dvd_modified} {
					set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon warning \
							-message "DVD modifiziert, aber nicht gespeichert!\n\nDie DVD wird in DVDShrink geladen mit\nden auf der Disk zurzeit gespeicherten Dateien.\n\nFortfahren?" \
							-parent . -type yesno -default no -skip dvdshrink_plugin_dvd_modified]
					if {$r == "no"} {
						return
					}
				}
				catch {exec -- $filename $::dvddir &}
			} else {
				catch {exec -- $filename &}
			}
		} else {
			bell
		}
	} else {
		bell
	}
}

proc clear_analysis_cache {} {
	if {[catch {set common_appdata [registry get \
			"HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders" \
			"Common AppData"]}]} {
		set common_appdata "C:\\Documents and Settings\\All Users\\Application Data"
	}
	set dir [file join $common_appdata "DVD Shrink"]
	if {! [catch {set files [glob -nocomplain -directory [file join $common_appdata "DVD Shrink"] -- *.*]}]} {
		foreach f $files {
			if {[file size $f] != 4} {
				catch {file delete $f}
			}
		}
	}
}

proc fix_50_titles {{quiet false}} {
	if {! [test_version 8.1 $::plugins::dvdshrink::plugintitle]} {return false}
	if {! [info exists ::dvddir]} {
		bell
		return false
	}
	set maxvts $::pgcs(numvts)
	if {!$quiet} {
		if {$maxvts < 50} {
		set txt "Diese Funktion ist bestimmt, um ein groes Problem im First-Play PGC zu korrigieren,"
		set txt "$txt\nbei den Kompilationen, die mit DVDShrink im Modus \"Eigene DVD\" erstellt wurden,"
		set txt "$txt\nwenn dort 50 oder mehr Titel in der Kompilation sind." 
		set txt "$txt\nDa es nur $maxvts Titel in der gegenwrtigen DVD gibt, brauchen Sie diese Funktion" 
		set txt "$txt\nnicht anzuwenden."
			tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon info -message $txt \
					-parent .
			return false
		}
	
		set txt "Diese Funktion korrigiert ein groes Problem im First-Play PGC, bei Kompilationen die mit DVDShrink"
	  set txt "$txt\nim Modus \"Eigene DVD\" erstellt wurden, wenn es 50 oder mehr Titel in der Kompilation gibt."
	  set txt "$txt\n\nEs wird empfohlen, diese Funktion anzuwenden, auer wenn Sie ein Men mit dem TitleWriter erstellen wollen."
	  set txt "$txt\n\nBemerken Sie auch, dass diese Funktion mit \"DVDShrink - Navigationstasten\" inkompatibel ist,"
	  set txt "$txt\n(aber Sie knnen \"Authoring nochmal !\" benutzen, ohne diese Funktion zuerst anzuwenden.)"
	  set txt "$txt\n"
	  set txt "$txt\nFortfahren ?"
		set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon question -message $txt \
				-parent . -type yesno -default yes -skip dvdshrink_plugin_fix50_info]
		if {$r == "no"} {return false}
	}

	if {$::menus(0,numlus)>0} {
		set txt "WARNUNG: Es gibt schon ein VMGM Men!"
		set txt "$txt\nWenn Sie fortfahren, wird es durch ein neues ersetzt!"
		set txt "$txt\n"
		set txt "$txt\nTrotzdem fortfahren ?"
		set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon warning -message $txt \
				-parent . -type yesno -default yes -skip dvdshrink_plugin_replace_vmgm]
	}
	if {$r == "no"} {
		return false
	}

	### create or replace the VMGM menu

	# FP-PGC - Init GPRMs and jump to title 1
	# 1  (JumpTT) Jump to VMGM PGC 2
	set ::pgcs(0,0) [list \
		{* pre} \
		{30 06 00 02 00 C0 00 00} \
		{* post} \
		{* cell} ]

	# Create a VMGM menu with 1 dummy PGC
	set ::raw_sectors(0,VMGM_PGCI_UT) \
			[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]

	::readdvd::parse_PGCI_UT ::raw_sectors(0,VMGM_PGCI_UT) 0 ::menus
	::main::init_tree_menu
	::main::display_pgc menu 0 1 1

	set cmds {{* pre} {61 00 00 00 00 84 00 00}}
	lappend cmds "71 A0 00 00 00 00 00 [format %02X $maxvts]"
	lappend cmds {20 04 00 00 00 00 00 02}
	lappend cmds {* post}
	lappend cmds {* cell}
	set ::menus(0,1,1) $cmds

	# create a new PGC
	::utils::new_pgc true false

	# adds another PGC if needed
	if {$maxvts > 64} {
		::utils::new_pgc true false
		set cmds {{* pre} {20 D4 00 00 00 3F 00 03}}
		set gotoline 66
		set numgotos 63
	} else {
		set cmds {{* pre}}
		set gotoline [expr $maxvts+1]
		set numgotos [expr $maxvts-1]
	}
	for {set i 1} {$i <= $numgotos} {incr i} {
		set cmd "00 A1 00 00 00 [format %02X $i] 00 [format %02X $gotoline]"
		lappend cmds $cmd
		incr gotoline
	}
	incr numgotos
	for {set i 1} {$i <= $numgotos} {incr i} {
		set cmd "30 02 00 00 00 [format %02X $i] 00 00"
		lappend cmds $cmd
	}
	lappend cmds {* post}
	lappend cmds {* cell}
	set ::menus(0,1,2) $cmds

	if {$maxvts > 64} {
		set cmds {{* pre}}
		set remain [expr $maxvts-64]
		if {$remain > 64} {set remain 64}
		set gotoline [expr $remain+1]
		set numgotos [expr $remain-1]

		for {set i 1} {$i <= $numgotos} {incr i} {
			set cmd "00 A1 00 00 00 [format %02X [expr $i+64]] 00 [format %02X $gotoline]"
			lappend cmds $cmd
			incr gotoline
		}
		incr numgotos
		for {set i 1} {$i <= $numgotos} {incr i} {
			set cmd "30 02 00 00 00 [format %02X [expr $i+64]] 00 00"
			lappend cmds $cmd
		}
		lappend cmds {* post}
		lappend cmds {* cell}
		set ::menus(0,1,3) $cmds
	}

	### util done.  Refresh listbox.
	::main::display_pgc pgc 0 0
	::main::dvd_modified true false

	# if there is an old VMGM VOB, delete it or move it to backup
	::utils::delete_vob_with_confirm "VIDEO_TS.VOB" true

	return true
}

# Redo authoring of a DVDShrink compilation in re-author mode.
# This method is probably the most standard compliant, and is my prefered one.
# Support for BOVs to use the <- and -> buttons.
# The DVDShrink bugs when there are 50 or more titles is also fixed.
proc best_method {} {
	if {! [test_version 8.1 $::plugins::dvdshrink::plugintitle]} {return false}
	if {! [info exists ::dvddir]} {
		bell
		return false
	}

	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]	;# last PGC in domain
	if {$numtitles < 2} {
		bell
		return false
	}
	if {$numtitles > 99} {
		set txt "WARNUNG: Es sind $numtitles Titel in dieser DVD !"
		set txt "$txt\n\nDer DVD-Videostandard beschrnkt die Anzahl von Titeln auf 99."
		set txt "$txt\nJedoch untersttzt diese Funktion bis zu 126 Titel, aber wenn"
		set txt "$txt\nSie fortfahren wird eine unzulssige DVD wird produziert !"
		if {$numtitles > 126} {
			set txt "$txt\n\nDa die Kompilation jedenfalls mehr als 126 Titel enthlt,"
			set txt "$txt\nwerden nur die ersten 126 Titel in Betracht gezogen."
			set txt "$txt\nSie werden nicht im Stande sein, die nachfolgenden Titel abzuspielen"
			set txt "$txt\nund wahrscheinlich auch nicht die Titel 100 bis 126 !"
		}
		set txt "$txt\n\nEs wird empfohlen, den Prozess abzubrechen und nochmals eine"
		set txt "$txt\nKompilation mit maximal 99 Titeln zu erstellen, oder einige Titel oder"
		set txt "$txt\nTitelSets mit PgcEdit vor der Verwendung dieser Funktion zu lschen."
		tk_messageBox -title "DVDShrink Plugin" -message $txt -icon warning
	}

	set ::plugins::dvdshrink::enable_next 1
	if {[get_int ::pgcs(1,1,header) 0X9C 2] != 1} {
		set ::plugins::dvdshrink::enable_next 0
	}
	set ::plugins::dvdshrink::enable_prev 1
	if {[get_int ::pgcs(1,1,header) 0X9E 2] != 1} {
		set ::plugins::dvdshrink::enable_prev 0
	}
	set ::plugins::dvdshrink::enable_goup 1
	if {[get_int ::pgcs(1,1,header) 0XA0 2] != 1} {
		set ::plugins::dvdshrink::enable_goup 0
	}
	set ::plugins::dvdshrink::last_ptt 0
	if {[lsearch [string toupper [get_type_cmds pre ::pgcs(1,1)]] {30 08 00 03 01 C0 00 00}] >= 0} {
		set ::plugins::dvdshrink::last_ptt 1
	}

	set ::plugins::dvdshrink::enable_arrows 2
	set ::plugins::dvdshrink::swap_arrows 0
	if {[info exists ::buttons::butsinfo(pgc,1,1,1,1,2,command)] && \
		[string toupper $::buttons::butsinfo(pgc,1,1,1,1,2,command)] == {30 08 00 03 01 C0 00 00}} {
		set ::plugins::dvdshrink::swap_arrows 1
	}

	set ::plugins::dvdshrink::redo_root 0
	if {$::menus(1,numlus)==1} {
		if {[string toupper [lindex $::menus(1,1,1) 2]] == "30 06 00 01 00 C0 00 00"} {
			set ::plugins::dvdshrink::redo_root 1
		}
		if {[string toupper [lindex $::menus(1,1,1) 2]] == "30 06 00 02 00 C0 00 00"} {
			set ::plugins::dvdshrink::redo_root 2
		}
	}
	set ::plugins::dvdshrink::redo_title 0
	if {$::menus(0,numlus)==1 && ($::menus(0,1,numpgcs)==3 || $::menus(0,1,numpgcs)==4)} {
		if {$::menus(0,1,1,masktext)=="TitleM"} {
			set ::plugins::dvdshrink::redo_title 1
		}
		if {$::menus(0,1,2,masktext)=="TitleM"} {
			set ::plugins::dvdshrink::redo_title 2
		}
	}
	set ::plugins::dvdshrink::redo_exit 0
	if {[lsearch [get_type_cmds post ::pgcs([lindex $::pgcs(Title,$numtitles) 0],[lindex $::pgcs(Title,$numtitles) 1])] {30 01 00 00 00 00 00 00}] >= 0} {
		set ::plugins::dvdshrink::redo_exit 1
	}

	### GUI
	set w .dvdshrinkredo
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w "$::plugins::dvdshrink::plugintitle:  Erweitern und Korrigieren einer DVDShrink-Kompilationen (Eigene DVD)"
	if {$::config(tk) != "aqua"} {wm transient $w .}
	wm protocol $w WM_DELETE_WINDOW {set ::plugins::dvdshrink::redo_done 0}
	bind $w <Escape> "set ::plugins::dvdshrink::redo_done 0"
	bind $w <Return> "set ::plugins::dvdshrink::redo_done 1"

	set txt "WARNUNG: Diese Funktion sollte nur auf eine nicht modifizierte DVDShrink-Kompilation \"Eigene DVD\" angewendet werden!"
	set txt "$txt\n(Jedoch knnen Sie diese Funktion mehrere Male auf derselben DVD sicher anwenden, um die Optionen zu testen und Sie knnen"
	set txt "$txt\nauch auf Ihr eigenes Risiko es wahrscheinlich auch auf jeder Kompilation mit sequentiellen Titeln ohne Mens anwenden.)"
	label $w.lab -text $txt -justify left
	pack  $w.lab -padx 4 -anchor w

	frame $w.f3 -bd 2 -relief groove
	pack  $w.f3 -padx 4 -pady 1 -fill x
	checkbutton $w.f3.b3 -text "Den Button (Next Chapter) einschalten, um whrend der Wiedergabe des letzten Kapitels eines Titels zum nchsten Titel zu springen.\n(Das Verhalten whrend der Wiedergabe des letzten Kapitels von dem letzten Titel hngt von der Option \"Beenden\" ganz unten ab.)" \
			-variable ::plugins::dvdshrink::enable_next -justify left -pady 1 -bd 0
	pack  $w.f3.b3 -anchor w
	checkbutton $w.f3.b2 -text "Den Button (Previous Chapter) einschalten, um whrend der Wiedergabe des ersten Kapitels zum vorherigen Titel zu springen\n(oder wenn das erste Kapitel des ersten Titels abgespielt wird, eine Schleife (Loop) zurck zum letzten Titel.)" \
			-variable ::plugins::dvdshrink::enable_prev -justify left -pady 1 -bd 0
	pack  $w.f3.b2 -anchor w
	checkbutton $w.f3.b1 -text "Den Button (Return / GoUp) einschalten, um immer zum vorherigen Titel zu springen\n(oder wenn der erste Titel abgespielt wird, eine Schleife (Loop) zurck zum letzten Titel.)" \
			-variable ::plugins::dvdshrink::enable_goup -justify left -pady 1 -bd 0
	pack  $w.f3.b1 -anchor w
	checkbutton $w.f3.b4 -text "ndern des Verhaltens der Button-Optionen oben (Previous Chapter und Return / GoUp), um zurck\nzum letzten Kapitel des vorherigen Titels zu springen (statt zum Anfang des vorherigen Titels.)" \
			-variable ::plugins::dvdshrink::last_ptt -justify left -pady 1 -bd 0
	pack  $w.f3.b4 -anchor w
	label $w.f3.lab -text "Hinweis: Einige schlechte Software-Player ignorieren einige jener Optionen oder behandeln sie falsch, oder sie machen den Suchbalken\nunbrauchbar, wenn mindestens eine jener Optionen angehakt werden.  (Standardeinstellung DVDShrink: alle Optionen nicht angehakt)" -justify left
	pack  $w.f3.lab -anchor w

	frame $w.f4 -bd 2 -relief groove
	pack  $w.f4 -padx 4 -pady 1 -fill x
	radiobutton $w.f4.b1 -text "Einschalten der Pfeil-Tasten (Steuerkreuz): LINKS -> Vorheriger Titel,  RECHTS -> Nchster Titel,  HOCH -> Letztes Kapitel des vorherigen Titel,\nRUNTER -> Nchster Titel (oder Beenden am Ende der DVD)    Siehe die Option \"Beenden\" unten, ENTER -> Neustart des aktuellen Titel." \
			-variable ::plugins::dvdshrink::enable_arrows -value 1 -pady 0 -bd 0 -justify left
	pack  $w.f4.b1 -anchor w
	checkbutton $w.f4.c1 -text "Tauschen der Funktionen der Buttons (LINKS und HOCH) und der Buttons (RECHTS und RUNTER)." \
			-variable ::plugins::dvdshrink::swap_arrows -pady 0 -bd 0 -justify left
	pack  $w.f4.c1 -anchor w -padx 18
	radiobutton $w.f4.b2 -text "Ausschalten der Pfeil-Tasten (Steuerkreuz) und entfernen aller BOVs.     (Standardeinstellung DVDShrink)" \
			-variable ::plugins::dvdshrink::enable_arrows -value 0 -pady 0 -bd 0
	pack  $w.f4.b2 -anchor w
	radiobutton $w.f4.b3 -text "Diese Einstellung nicht ndern, da sie viel Zeit bentigt, wenn die DVD gespeichert wird." \
			-variable ::plugins::dvdshrink::enable_arrows -value 2 -pady 0 -bd 0
	pack  $w.f4.b3 -anchor w

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 1 -fill x
	label $w.f1.lab -text "Wenn der TITLE MEN Button gedrckt wird..." -pady 0
	pack  $w.f1.lab -anchor w
	radiobutton $w.f1.b1 -text "springe zum Anfang des nchsten Titels (oder wenn der letzte Titel abgespielt wird, springe zurck zum ersten Titel.)" \
			-variable ::plugins::dvdshrink::redo_title -value 1 -pady 0 -bd 0
	pack  $w.f1.b1 -anchor w
	radiobutton $w.f1.b2 -text "springe zum Anfang des vorherigen Titels (oder wenn der erste Titel abgespielt wird, eine Schleife zurck zum letzten Titel.)" \
			-variable ::plugins::dvdshrink::redo_title -value 2 -pady 0 -bd 0
	pack  $w.f1.b2 -anchor w
	radiobutton $w.f1.b0 -text "nichts tun.  (Standardeinstellung von DVDShrink)" \
			-variable ::plugins::dvdshrink::redo_title -value 0 -pady 0 -bd 0
	pack  $w.f1.b0 -anchor w

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 1 -fill x
	label $w.f2.lab -text "Wenn der ROOT MEN Button gedrckt wird..." -pady 0
	pack  $w.f2.lab -anchor w
	radiobutton $w.f2.b1 -text "springe zum Anfang des nchsten Titels (oder wenn der letzte Titel abgespielt wird, springe zurck zum ersten Titel.)" \
			-variable ::plugins::dvdshrink::redo_root -value 1 -pady 0 -bd 0
	pack  $w.f2.b1 -anchor w
	radiobutton $w.f2.b2 -text "springe zum Anfang des vorherigen Titels (oder wenn der erste Titel abgespielt wird, eine Schleife zurck zum letzten Titel.)" \
			-variable ::plugins::dvdshrink::redo_root -value 2 -pady 0 -bd 0
	pack  $w.f2.b2 -anchor w
	radiobutton $w.f2.b0 -text "nichts tun.  (Standardeinstellung von DVDShrink)" \
			-variable ::plugins::dvdshrink::redo_root -value 0 -pady 0 -bd 0
	pack  $w.f2.b0 -anchor w

	frame $w.f5 -bd 2 -relief groove
	pack  $w.f5 -padx 4 -pady 1 -fill x
	label $w.f5.lab -text "Option Beenden: Wenn der letzte Titel fertig abgespielt wurde (oder wenn der Button NEXT CHAPTER gedrckt wird, whrend der Wiedergabe des\nletzten Kapitels von dem letzten Titel)..." -pady 0 -justify left
	pack  $w.f5.lab -anchor w
	radiobutton $w.f5.b1 -text "eine Schleife (Loop) zurck zum Anfang des ersten Titels." \
			-variable ::plugins::dvdshrink::redo_exit -value 0 -pady 0 -bd 0
	pack  $w.f5.b1 -anchor w
	radiobutton $w.f5.b0 -text "Beenden der DVD-Wiedergabe.  (Standardeinstellung von DVDShrink)" \
			-variable ::plugins::dvdshrink::redo_exit -value 1 -pady 0 -bd 0
	pack  $w.f5.b0 -anchor w

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 2 -fill x
	button $w.fb.b5 -text "Einstellungen laden" -width 18 -command {
		set ::plugins::dvdshrink::enable_goup 1
		set ::plugins::dvdshrink::enable_prev 1
		set ::plugins::dvdshrink::enable_next 1
		set ::plugins::dvdshrink::last_ptt    0
		set ::plugins::dvdshrink::redo_title  1
		set ::plugins::dvdshrink::redo_root   1
		set ::plugins::dvdshrink::redo_exit   1
		set ::plugins::dvdshrink::enable_arrows 1
		set ::plugins::dvdshrink::swap_arrows 0
		set filename [file join $::config(config_directory) dvdshrink_plugin.cfg]
		if {[file exists $filename]} {
			catch {source $filename}
		}
	}
	pack   $w.fb.b5 -side left -padx 2
	button $w.fb.b6 -text "Einstellungen speichern" -width 20 -command {
		set txt "# DVDShrink plugin v$::plugins::dvdshrink::version defaults file"
		set txt "$txt\n\nset ::plugins::dvdshrink::enable_goup $::plugins::dvdshrink::enable_goup"
		set txt "$txt\nset ::plugins::dvdshrink::enable_prev $::plugins::dvdshrink::enable_prev"
		set txt "$txt\nset ::plugins::dvdshrink::enable_next $::plugins::dvdshrink::enable_next"
		set txt "$txt\nset ::plugins::dvdshrink::last_ptt $::plugins::dvdshrink::last_ptt"
		set txt "$txt\nset ::plugins::dvdshrink::redo_title $::plugins::dvdshrink::redo_title"
		set txt "$txt\nset ::plugins::dvdshrink::redo_root $::plugins::dvdshrink::redo_root"
		set txt "$txt\nset ::plugins::dvdshrink::redo_exit $::plugins::dvdshrink::redo_exit"
		set txt "$txt\nset ::plugins::dvdshrink::enable_arrows $::plugins::dvdshrink::enable_arrows\n"
		set txt "$txt\nset ::plugins::dvdshrink::swap_arrows $::plugins::dvdshrink::swap_arrows\n"
		saveFile [file join $::config(config_directory) dvdshrink_plugin.cfg] auto $txt
	}
	pack   $w.fb.b6 -side left -padx 2

	button $w.fb.cancel -text "Abbrechen" -width 10 -command {set ::plugins::dvdshrink::redo_done 0}
	pack   $w.fb.cancel -side right -padx 2
	button $w.fb.ok -text "OK" -width 10 -command {set ::plugins::dvdshrink::redo_done 1}
	pack   $w.fb.ok -side right -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	focus $w
	grab set $w

	tkwait variable ::plugins::dvdshrink::redo_done
	grab release $w
	destroy $w
	if {!$::plugins::dvdshrink::redo_done} return

	if {! $::plugins::dvdshrink::enable_prev && ! $::plugins::dvdshrink::enable_goup} {
		set ::plugins::dvdshrink::last_ptt 0
	}

	show_msg "DVDShrink Plugin: Anwenden der Erweiterungen..."
	set ::gprm_labels(0) "titleCtrl"
	set ::gprm_labels(1) "lastPttCtrl"
	set ::gprm_labels(2) "HlBtnBackup"

	### create or replace the VMGM menu

	::main::display_pgc pgc 0 0
	show_msg "Bearbeite das VMGM Men - Bitte warten..." 0 $numtitles

	# FP-PGC
	# 1  Set gprm(0) =(mov) 1
	# 2  Set gprm(1) =(mov) 0 
	# 3  Set gprm(2) =(mov) 1024 
	# 4  (SetHL_BTN) Set Highlighted Button =(mov) 1024 (button 1) 
	# 5  Jump to title 1
	set ::pgcs(0,0) [list {* pre} {71 00 00 00 00 01 00 00} {71 00 00 01 00 00 00 00} \
		{71 00 00 02 04 00 00 00} {56 00 00 00 04 00 00 00} {30 02 00 00 00 01 00 00} \
		{* post} {* cell}]

	# Create a VMGM menu
	set ::raw_sectors(0,VMGM_PGCI_UT) \
			[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]
	::readdvd::parse_PGCI_UT ::raw_sectors(0,VMGM_PGCI_UT) 0 ::menus
	::main::init_tree_menu
	::main::display_pgc menu 0 1 1

	# VMGM PGC 1: jump to next title
	# 1  Set gprm(0) &=(and) 127 
	# 2  Set gprm(0) +=(add) 1
	# 3  if ( gprm(0) > last_title ) then { Set gprm(0) =(mov) 1 } 
	# 4  (SetHL_BTN) Set Highlighted Button =(mov) gprm(2) 
	# 5  LinkPGCN PGC 4
	set cmds {{* pre} {79 00 00 00 00 7F 00 00} {73 00 00 00 00 01 00 00}} 
	lappend cmds "71 D0 00 00 00 01 00 [format %02X $numtitles]"
	lappend cmds {46 00 00 00 00 02 00 00}
	lappend cmds {20 04 00 00 00 00 00 04}
	lappend cmds {* post}
	lappend cmds {* cell}
	set ::menus(0,1,1) $cmds
	set ::menus_comments(0,1,1) "Sprung zum nchsten Titel"


	# VMGM PGC 2: jump to previous title
	# 1  Set gprm(0) &=(and) 127 
	# 2  Set gprm(0) -=(sub) 1 
	# 3  if ( gprm(0) < 1 ) then { Set gprm(0) =(mov) last_title } 
	# 4  (SetHL_BTN) Set Highlighted Button =(mov) gprm(2) 
	# 5  LinkPGCN PGC 4
	::utils::new_pgc true false
	set cmds {{* pre} {79 00 00 00 00 7F 00 00} {74 00 00 00 00 01 00 00}}
	lappend cmds "71 F0 00 00 00 [format %02X $numtitles] 00 01"
	lappend cmds {46 00 00 00 00 02 00 00}
	lappend cmds {20 04 00 00 00 00 00 04}
	lappend cmds {* post}
	lappend cmds {* cell}
	set ::menus(0,1,2) $cmds
	set ::menus_comments(0,1,2) "Sprung zum vorherigen Titel"

	# VMGM PGC 3: jump to last chapter of previous title
	# 1  Set gprm(1) =(mov) 1
	# 2  LinkPGCN PGC 2 
	::utils::new_pgc true false
	set cmds {{* pre} {71 00 00 01 00 01 00 00} {20 04 00 00 00 00 00 02} {* post} {* cell}}
	set ::menus(0,1,3) $cmds
	set ::menus_comments(0,1,3) "Sprung zum vorherigen Kapitel"

	# Change the Title menu according to the user wishes
	switch $::plugins::dvdshrink::redo_title {
		1 {
			# jump to next title: nothing to do
			::utils::set_menu_type 0 1 1 2 
		}
		2 {
			# jump to prev title: Title menu is VMGM PGC 2
			::utils::set_menu_type 0 1 2 2 
		}
		0 {
			# do nothing: remove Title menu ID
			::utils::set_menu_type 0 1 1 0 
			::utils::set_menu_type 0 1 2 0 
		}
	}

	# VMGM PGC 4: jump to title pointed to by GPRM(0)
	::utils::new_pgc true false
	set cmds {{* pre}}
	set ::menus_comments(0,1,4) "Sprung zum Titel in GPRM 0"
	# adds another PGC if needed (when there are more than 63 titles)
	if {$numtitles > 63} {
		::utils::new_pgc true false
		set ::menus_comments(0,1,5) "dasselbe, fortgesetzt"
		lappend cmds {20 D4 00 00 00 3F 00 05}	;# if ( gprm(0) > 63 ) then { LinkPGCN PGC 5 } 
		set gotoline 66
		set numgotos 63
	} else {
		set gotoline [expr $numtitles+2]
		set numgotos $numtitles
	}
	for {set i 1} {$i <= $numgotos} {incr i} {
		# if ( gprm(0) == title ) then { Goto line gotoline }
		set cmd "00 A1 00 00 00 [format %02X $i] 00 [format %02X $gotoline]"
		lappend cmds $cmd
		incr gotoline
	}
	lappend cmds "30 06 00 00 00 00 00 00"	;# (JumpSS) Jump to First Play PGC
	for {set i 1} {$i <= $numgotos} {incr i} {
		# (JumpTT) Jump to Title title 
		set cmd "30 02 00 00 00 [format %02X $i] 00 00"
		lappend cmds $cmd
	}
	lappend cmds {* post}
	lappend cmds {* cell}
	set ::menus(0,1,4) $cmds

	if {$numtitles > 63} {
		::main::display_pgc menu 0 1 5
		set remain [expr $numtitles-63]
		if {$remain > 63} {set remain 63}
		set gotoline [expr $remain+3]
		set numgotos $remain

		set cmds {{* pre} {20 F4 00 00 00 40 00 04}}	;# 1  if ( gprm(0) < 64 ) then { LinkPGCN PGC 4 } 
		for {set i 1} {$i <= $numgotos} {incr i} {
			# if ( gprm(0) == title ) then { Goto line gotoline }
			set cmd "00 A1 00 00 00 [format %02X [expr $i+63]] 00 [format %02X $gotoline]"
			lappend cmds $cmd
			incr gotoline
		}
		lappend cmds "30 06 00 00 00 00 00 00"	;# (JumpSS) Jump to First Play PGC 
		for {set i 1} {$i <= $numgotos} {incr i} {
			# (JumpTT) Jump to Title title 
			set cmd "30 02 00 00 00 [format %02X [expr $i+63]] 00 00"
			lappend cmds $cmd
		}
		lappend cmds {* post}
		lappend cmds {* cell}
		set ::menus(0,1,5) $cmds
	}

	# modify the commands of the titles
	# *** pre commands
	# 1  if ( gprm(0) < 128 ) then { Goto line 4 } 
	# 2  Set gprm(0) &=(and) 127 
	# 3  (CallSS) Call the VMGM PGC 2, resume cell 1 ;# or PGC 3 if the option to return to the prev chapter has been selected
	# 4  NOP 
	# 5  Set gprm(0) =(mov) 1 
	# 6  Set gprm(0) |=(or) 128 
	# 7  Set gprm(2) =(mov) sprm(8:Highlighted button number) 
	# 8  (SetHL_BTN) Set Highlighted Button =(mov) 1024 (button 1) 
	# 9  if ( gprm(1) == 0 ) then { Break } 
	# 10  Set gprm(1) =(mov) 0 
	# 11 LinkPGN Program last_chapter 
	# *** post commands
	# 1  Set gprm(0) &=(and) 127 
	# 2  (CallSS) Call the VMGM PGC 1, resume cell 1	;# or Exit
	set cmds {{* pre}}
	lappend cmds "00 F1 00 00 00 80 00 04"
	lappend cmds "79 00 00 00 00 7F 00 00"
	if {! $::plugins::dvdshrink::last_ptt} {
		lappend cmds "30 08 00 02 01 C0 00 00"
	} else {
		lappend cmds "30 08 00 03 01 C0 00 00"
	}
	lappend cmds "00 00 00 00 00 00 00 00"

	lappend cmds "00 00 00 00 00 00 00 00"  ;# room for original SetSTN commands
	lappend cmds "00 00 00 00 00 00 00 00"

	lappend cmds "71 00 00 00 00 01 00 00"	;# current_title is set later
	lappend cmds "7A 00 00 00 00 80 00 00"
	lappend cmds "61 00 00 02 00 88 00 00"
	lappend cmds "56 00 00 00 04 00 00 00"
	lappend cmds "00 A2 00 01 00 00 00 00"
	lappend cmds "71 00 00 01 00 00 00 00"
	lappend cmds "20 06 00 00 00 00 00 01"	;# last PG is set later
	lappend cmds {* post}
	lappend cmds "79 00 00 00 00 7F 00 00"
	lappend cmds "30 08 00 01 01 C0 00 00"	;# possibly changed later to Exit
	lappend cmds {* cell}
	for {set i 1} {$i <= $numtitles} {incr i} {
		set vts [lindex $::pgcs(Title,$i) 0]
		set pgc [lindex $::pgcs(Title,$i) 1]
		::main::display_pgc pgc $vts $pgc

		show_msg "Bearbeite Titel $i - Bitte warten..." [expr $i+1] $numtitles
		lset cmds 7 "71 00 00 00 00 [format %02X $i] 00 00"	;# Set gprm(0) =(mov) current_title

		# copy the original SetSTN commands
	    lset cmds 5 "00 00 00 00 00 00 00 00"
	    lset cmds 6 "00 00 00 00 00 00 00 00"
		set setstncmds [lsearch -all -inline -regexp [get_type_cmds pre ::pgcs($vts,$pgc)] {[45]1 00 .. .. .. .. .. ..}]
		if {[llength $setstncmds] > 0} {
		    lset cmds 5 [lindex $setstncmds 0]
		}
		if {[llength $setstncmds] > 1} {
		    lset cmds 6 [lindex $setstncmds 1]
		}

		# compute last PTT
		set programmap $::pgcs($vts,$pgc,programmap)
		set numprograms [string length $programmap]
		set cellpbtable $::pgcs($vts,$pgc,cellpbtable)
		set pg $::pgcs($vts,$pgc,numprograms)
		while {$pg > 0} {
			set pgcell [get_int programmap [expr $pg-1] 1]
			if {$::pgcs($vts,$pgc,numcells) < $pgcell || $pgcell == 0} {
				incr pg -1
				continue
			}
			if {$pgcell == $::pgcs($vts,$pgc,numcells)} {
				set idx [expr (($pgcell-1)*24)+4]
				set bcdpart [string range $cellpbtable $idx [expr $idx+2]]
				if {$bcdpart == "\000\000\000"} {
					incr pg -1
					continue
				}
			}
			break
		}
		if {$pg < 1} {set pg 1}
		lset cmds 13 "20 06 00 00 00 00 00 [format %02X $pg]"

		# exit at end of last title if the user has requested so
		if {$::plugins::dvdshrink::redo_exit && $i == $numtitles} {
			lset cmds 16 "30 01 00 00 00 00 00 00"	;# Exit
		}
		set ::pgcs($vts,$pgc) $cmds

		# Modify the prev/next/goupPGCN links
		if {$::plugins::dvdshrink::enable_next} {
			set_int ::pgcs($vts,$pgc,header) 0X9C $pgc 2
		} else {
			set_int ::pgcs($vts,$pgc,header) 0X9C 0 2
		}
		if {$::plugins::dvdshrink::enable_prev} {
			set_int ::pgcs($vts,$pgc,header) 0X9E $pgc 2
		} else {
			set_int ::pgcs($vts,$pgc,header) 0X9E 0 2
		}
		if {$::plugins::dvdshrink::enable_goup} {
			set_int ::pgcs($vts,$pgc,header) 0XA0 $pgc 2
		} else {
			set_int ::pgcs($vts,$pgc,header) 0XA0 0 2
		}
		# clear important PUOs
		set_int ::pgcs($vts,$pgc,header) 0X08 \
				[expr [get_int ::pgcs($vts,$pgc,header) 0X08 4] & 0x1FDF30D] 4

		# Create a Root menu with 1 dummy PGC
		if {$::menus($vts,numlus)>0} {
			::utils::remove_menu true
		}
		switch $::plugins::dvdshrink::redo_root {
			1 {
				# jump to next title
				::utils::new_root_menu $vts true
				# 1  Set gprm(0) &=(and) 127 
				# 2  (SetHL_BTN) Set Highlighted Button =(mov) gprm(2) 
				# 3  (JumpSS) Jump to VMGM PGC 1
				set ::menus($vts,1,1) {"* pre" "79 00 00 00 00 7F 00 00" \
						"46 00 00 00 00 02 00 00" "30 06 00 01 00 C0 00 00" "* post" "* cell"}
			}
			2 {
				# jump to prev title
				::utils::new_root_menu $vts true
				# 1  Set gprm(0) &=(and) 127 
				# 2  (SetHL_BTN) Set Highlighted Button =(mov) gprm(2) 
				# 3  (JumpSS) Jump to VMGM PGC 2
				set ::menus($vts,1,1) {"* pre" "79 00 00 00 00 7F 00 00" \
						"46 00 00 00 00 02 00 00" "30 06 00 02 00 C0 00 00" "* post" "* cell"}
			}
			default {
				# do nothing: do not create the root menu
			}
		}
	}

	if {$::plugins::dvdshrink::enable_arrows != 2} {
		show_msg "Initialisieren der BOVs - Bitte warten..." 100
		if {$::plugins::dvdshrink::swap_arrows} {
			set butcmds {{20 06 00 00 00 00 00 01} \
				{30 08 00 03 01 C0 00 00} {20 01 00 00 00 00 00 0D} \
				{30 08 00 02 01 C0 00 00} {30 08 00 01 01 C0 00 00}}
		} else {
			set butcmds {{20 06 00 00 00 00 00 01} \
				{30 08 00 02 01 C0 00 00} {30 08 00 01 01 C0 00 00} \
				{30 08 00 03 01 C0 00 00} {20 01 00 00 00 00 00 0D}}
		}
		set ::menued::scan_done 1
		set ::buttons::butsinfo(pgc,multiplesets) false
		for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
			set totalbuttons 0
			unset -nocomplain numcellids
			unset -nocomplain navpacks
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set cellpbtable  $::pgcs($vts,$pgc,cellpbtable)
				set cellpostable $::pgcs($vts,$pgc,cellpostable)
				set numvobids 0
				for {set cell 0} {$cell < $::pgcs($vts,$pgc,numcells)} {incr cell} {
					incr totalbuttons 5
					set cellpbidx  [expr ($cell*24)+8]
					set cellposidx [expr $cell*4]
					set vobid  [get_int cellpostable $cellposidx 2]
					set cellid [get_int cellpostable [expr $cellposidx+3] 1]
					if {$vobid > $numvobids} {
						set numvobids $vobid
					}
					if {![info exists numcellids($vobid)]} {
						set numcellids($vobid) $cellid
					} elseif {$cellid > $numcellids($vobid)} {
						set numcellids($vobid) $cellid
					}
					if {![info exists navpacks($vobid,$cellid)]} {
						set navpacks($vobid,$cellid) [get_int cellpbtable $cellpbidx 4]
					}
				}
				set ::buttons::butsinfo(pgc,$vts,$vobid,numcellids) $numcellids($vobid)
				if {$::plugins::dvdshrink::enable_arrows == 1} {
					set ::pgcs($vts,$pgc,numbuttons) [expr $::pgcs($vts,$pgc,numcells)*5]
					if {[get_int ::pgcs($vts,$pgc,header) 0x1C 4] == 0} {
						set_int ::pgcs($vts,$pgc,header) 0x1C 0x9F1F1F1F 4
					}
					if {[get_int ::raw_sectors($vts,VTSI_MAT) 0x255 1] < 1} {
						set_int ::raw_sectors($vts,VTSI_MAT) 0x255 1 1
					}
				} else {
					set ::pgcs($vts,$pgc,numbuttons) 0
					if {[get_int ::pgcs($vts,$pgc,header) 0x1C 4] == 0x9F1F1F1F && \
							[get_int ::raw_sectors($vts,VTSI_MAT) 0x255 1] == 1 && \
							[get_int ::raw_sectors($vts,VTSI_MAT) 0x256 2] == 0 && \
							[get_int ::raw_sectors($vts,VTSI_MAT) 0x258 4] == 0} {
						set_int ::pgcs($vts,$pgc,header) 0x1C 0 4
						set_int ::raw_sectors($vts,VTSI_MAT) 0x255 0 1
					}
				}
			}
			set ::buttons::butsinfo(pgc,$vts,errorcode) 0
			set ::buttons::butsinfo(pgc,$vts,errormsg) OK
			set ::buttons::butsinfo(pgc,$vts,numvobids) $numvobids
			set ::buttons::butsinfo(pgc,$vts,scanned) 1
			set ::buttons::butsinfo(pgc,$vts,edited) 1

			if {$::plugins::dvdshrink::enable_arrows == 1} {
				set ::buttons::butsinfo(pgc,$vts,totalbuttons) $totalbuttons
			} else {
				set ::buttons::butsinfo(pgc,$vts,totalbuttons) 0
			}

			for {set vobid 1} {$vobid <= $numvobids} {incr vobid} {
				if {[info exists numcellids($vobid)]} {
					set ::buttons::butsinfo(pgc,$vts,$vobid,edited) 1
					set ::buttons::butsinfo(pgc,$vts,$vobid,numcellids) $numcellids($vobid)
					for {set cellid 1} {$cellid <= $numcellids($vobid)} {incr cellid} {
						if {[info exists navpacks($vobid,$cellid)]} {
							set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,edited) 1
							set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,navpack) $navpacks($vobid,$cellid)

							set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,hlstarttime) 0
							set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,forceactionbutn) 0
							set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numselectablebuts) 0
							set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,grouptype) 0
							if {$::plugins::dvdshrink::enable_arrows == 1} {
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numgroups) 1
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numbuts) 5
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,hlendtime) -1
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,selendtime) -1
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,forceselectbutn) 1
								foreach b {1 2 3 4 5} {
									set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,$b,command) [lindex $butcmds [expr $b-1]]
									set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,$b,colorscheme) 3
									set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,$b,coords) {0 0 - 1 1}
									set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,$b,udlr) [string map "$b 1" {4 5 2 3}]
									set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,$b,autoaction) 1
								}
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,1,autoaction) 0
							} else {
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numgroups) 0
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numbuts) 0
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,hlendtime) 0
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,selendtime) 0
								set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,forceselectbutn) 0
								foreach b {1 2 3 4 5} {
									set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,$b,colorscheme) 0
									set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,$b,coords) {0 0 - 0 0}
									set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,$b,autoaction) 0
									set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,$b,command) {00 00 00 00 00 00 00 00}
									set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,1,$b,udlr) {0 0 0 0}
								}
							}
						}
					}
				}
			}
		}
		set ::buttons::butsinfo(pgc,edited) 1
		if {$::plugins::dvdshrink::enable_arrows == 1} {
			set ::config(auto_launch_menubuttons_parser) true
		} else {
			set filename [file join $::backupdir "Men-Buttons.but"]
			if {[file exists $filename]} {
				catch {file delete $filename}
			}
		}
	}

	### done.  Refresh listbox.
	::main::init_tree_menu
	::main::display_pgc pgc 0 0
	::main::dvd_modified true false
	set ::pgc_comments_modified true
	show_msg "Fertig."

	# if there is an old VMGM VOB, delete it or move it to backup
	::utils::delete_vob_with_confirm "VIDEO_TS.VOB" true

	return true
}

proc region_free {} {
	if {! [info exists ::dvddir]} {bell; return}
	if {[get_int ::raw_sectors(0,VMGM_MAT) 0X23 1] != 0} {
		::main::dvd_modified true false
		set_int ::raw_sectors(0,VMGM_MAT) 0X23 0 1
	}
}



} ;# end of namespace ::plugins::dvdshrink
