#############################################################################
# puo_plugin.tcl Copyright 2005, 2006 r0lZ
# PgcEdit plugin to import, export and clear prohibited user operations.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "Prohibited User Operations" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.0 or more to use this plugin.
# It is better to use v 0.6.1 or more.
#############################################################################
# Version History:
#
# v 1.2 (September 7, 2005)
# - First public release
# v 1.3 (November 3, 2005)
# - It is now possible to selectively keep some PUOs when removing them from
#   the VOB files.
# v 1.4 (October 7, 2006)
# - Bug fixed.  A wrong error message was displayed in some rare cases, when
#   the end of a VOB file is padded with null packs.
#############################################################################



# Add Plugins -> PUO submenu
if {! [winfo exists .mainmenu.plugins.iepuos]} {
	.mainmenu.plugins add cascade -label "Verbotene Benutzeraktionen (PUOs)" -menu .mainmenu.plugins.iepuos
	menu .mainmenu.plugins.iepuos
	.mainmenu.plugins.iepuos add command -label "Exportieren aller IFO PUOs" -command {::plugins::iepuos::export_ifo_puos}
	.mainmenu.plugins.iepuos add command -label "Importieren aller IFO PUOs" -command {::plugins::iepuos::import_ifo_puos}
	.mainmenu.plugins.iepuos add separator
	.mainmenu.plugins.iepuos add command -label "Festlegen der PUOs in der aktuellen Domain"  -command {::plugins::iepuos::set_domain_puos}
	.mainmenu.plugins.iepuos add separator
	.mainmenu.plugins.iepuos add command -label "Entfernen aller IFO PUOs"  -command {::plugins::iepuos::clear_ifo_puos}
	.mainmenu.plugins.iepuos add command -label "Selektives Entfernen der VOB PUOs"  -command {::plugins::iepuos::clear_vob_puos}
	.mainmenu.plugins.iepuos add separator
	.mainmenu.plugins.iepuos add command -label "ber..." -command {
			tk_messageBox  -title "Plugin - Verbotene Benutzeraktionen" -icon info \
					-message "Plugin - Verbotene Benutzeraktionen fr PgcEdit 0.6.1 oder hher.\nVersion $::plugins::iepuos::version\nAutor: r0lZ\nFreeware!"
		}
}

namespace eval ::plugins {}
namespace eval ::plugins::iepuos {

set version "1.4"

# Export IFO PUOs
proc export_ifo_puos {} {
	if {![info exists ::dvddir]} {bell; return}
	set title "PUO Plugin - Exportieren aller IFO PUOs          "

	set filename [tk_getSaveFile -filetypes {{"Exportierte PUOs" {*.puo}}} \
		-parent . -initialdir $::config(import_initialdir) \
		-initialfile "PUOs.puo" -title $title]
	if {$filename == ""} {
		return
	}
	if {[string toupper [file extension $filename]] != ".PUO"} {
		set filename "${filename}.puo"
	}
	set ::config(import_initialdir) [file dirname $filename]

	set numpgcs 1

	set txt "# Exportierte Verbotene Benutzeraktionen\n"
	set txt "${txt}# DVD: \"$::dvddir\"\n"
	set txt "${txt}# DVD-TEXT General Name: \"[::utils::get_General_Name]\"\n"
	set txt "${txt}# Provider ID: \"[::utils::get_provider_ID]\"\n"
	set txt "${txt}# Anzahl der VTS: $::pgcs(numvts)\n\n"

	set txt "${txt}set plugin_version $::plugins::iepuos::version\n\n"

	set txt "${txt}# PUOs in der VMG_TT_SRPT (Tabelle von Titeln)\n"
	set tt_srpt_puos false
	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]
	set txt "${txt}set puos(numtitles) $numtitles\n"
	for {set i 1} {$i <= $numtitles} {incr i} {
		set puos [expr [get_int ::raw_sectors(0,VMG_TT_SRPT) [expr {8+(($i-1)*12)}] 1] & 3]
		set txt "${txt}set puos(title,$i) $puos\n"
		if {$puos != 0} {set tt_srpt_puos true}
	}
	set txt "${txt}set puos(title,exists) $tt_srpt_puos\n\n\n"

	set puos [get_int ::pgcs(0,0,header) 8 4]
	set txt "${txt}# [::main::pgc_label $::config(fixed_width_font) pgc 0 0]\n"
	set txt "${txt}set puos(pgc,0,0) $puos\t;# 0x[format %08X $puos]\n"

	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		set txt "${txt}\n"
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set puos [get_int ::menus($vts,$lu,$pgc,header) 8 4]
				set txt "${txt}# [::main::pgc_label $::config(fixed_width_font) menu $vts $lu $pgc]\n"
				set txt "${txt}set puos(menu,$vts,$lu,$pgc) $puos\t;# 0x[format %08X $puos]\n"
				incr numpgcs
			}
		}
		set txt "${txt}\n"
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set puos [get_int ::pgcs($vts,$pgc,header) 8 4]
				set txt "${txt}# [::main::pgc_label $::config(fixed_width_font) pgc $vts $pgc]\n"
				set txt "${txt}set puos(pgc,$vts,$pgc) $puos\t;# 0x[format %08X $puos]\n"
				incr numpgcs
			}
		}
	}
	set txt "${txt}\n# Anzahl von PGCs in der DVD: $numpgcs\n"

	if {! [saveFile $filename auto $txt]} {
		show_msg "Verbotene Benutzeraktionen erfolgreich exportiert von $numpgcs PGCs und $numtitles Titles in der VMG_TT_SRPT."
	} else {
		show_msg "FEHLER beim Exportieren der Verbotenen Benutzeraktionen von $numpgcs PGCs!"
	}
}

# Import IFO PUOs
proc import_ifo_puos {} {
	if {![info exists ::dvddir]} {bell; return}
	set title "PUO plugin - Importieren aller IFO PUOs          "

	set filename [tk_getOpenFile -filetypes {{"Exportierte PUOs" {*.puo}}} \
		-parent . -initialdir $::config(import_initialdir) \
		-initialfile "PUOs.puo" -title $title]
	if {$filename == ""} {
		return
	}
	set ::config(import_initialdir) [file dirname $filename]

	if {[catch {source $filename} errstr]} {
		show_msg "FEHLER beim Importieren der PUOs!"
		tk_messageBox -title $title -icon error -type ok \
			-message "Fehler \"$filename\":\n$errstr"
		return
	}

	# PUOs in VMG_TT_SRPT
	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]
	set puotitles 0
	for {set i 1} {$i <= $numtitles} {incr i} {
		if {[info exists puos(title,$i)]} {
			set tidx [expr {8+(($i-1)*12)}]
			set titletype [expr ([get_int ::raw_sectors(0,VMG_TT_SRPT) $tidx 1] & 0xFC) | $puos(title,$i)]
			set_int ::raw_sectors(0,VMG_TT_SRPT) $tidx $titletype 1
			incr puotitles
		}
	}

	# PUOs in PGCs
	set numpgcs 1
	set imported 0
	if {[info exists puos(pgc,0,0)]} {
		set_int ::pgcs(0,0,header) 8 $puos(pgc,0,0) 4
		incr imported
	}
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				incr numpgcs
				if {[info exists puos(menu,$vts,$lu,$pgc)]} {
					set_int ::menus($vts,$lu,$pgc,header) 8 $puos(menu,$vts,$lu,$pgc) 4
					incr imported
				}
			}
		}
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				incr numpgcs
				if {[info exists puos(pgc,$vts,$pgc)]} {
					set_int ::pgcs($vts,$pgc,header) 8 $puos(pgc,$vts,$pgc) 4
					incr imported
				}
			}
		}
	}
	show_msg "PUOs erfolgreich importiert in $imported / $numpgcs PGCs und $puotitles / $numtitles Titles."
	::main::dvd_modified true false
}

# Clear IFO PUOs
proc clear_ifo_puos {} {
	if {![info exists ::dvddir]} {bell; return}
	set title "PUO Plugin - Entfernen aller IFO PUOs          "

	if {! $::config(expert_mode) && [tk_messageBox -title $title -icon question -type okcancel \
			-default ok -message "Entfernen aller Verbotenen Benutzeraktionen in den IFOs?\nSind Sie sicher?"] == "cancel"} return

	# PUOs in VMG_TT_SRPT
	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]
	for {set i 1} {$i <= $numtitles} {incr i} {
		set tidx [expr {8+(($i-1)*12)}]
		set titletype [expr [get_int ::raw_sectors(0,VMG_TT_SRPT) $tidx 1] & 0xFC]
		set_int ::raw_sectors(0,VMG_TT_SRPT) $tidx $titletype 1
	}

	set numpgcs 1
	set_int ::pgcs(0,0,header) 8 0 4
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				incr numpgcs
				set_int ::menus($vts,$lu,$pgc,header) 8 0 4
			}
		}
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				incr numpgcs
				set_int ::pgcs($vts,$pgc,header) 8 0 4
			}
		}
	}
	show_msg "Verbotene Benutzeraktionen erfolgreich entfernt in $numpgcs PGCs und $numtitles Titles."
	::main::dvd_modified true false
}

# Set domain PUOs
proc set_domain_puos {} {
	if {![info exists ::dvddir]} {bell; return}
	set type [lindex $::editor_cmds_from 0]
	set vts [lindex [split [lindex $::editor_cmds_from 1] ,] 0]
	#if {$type == "pgc" && $vts == 0} {bell; return}

	set ::pgced::puos [expr 0xFFFFFFFF]
	if {$type == "pgc"} {
		if {$vts == 0} {
			set ::pgced::puos [get_int ::pgcs(0,0,header) 8 4]
		} else {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set ::pgced::puos [expr $::pgced::puos & [get_int ::pgcs($vts,$pgc,header) 8 4]]
			} 
		}
	} else {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set ::pgced::puos [expr $::pgced::puos & [get_int ::menus($vts,$lu,$pgc,header) 8 4]]
			}
		}
	}
	if {[catch {set puos [::pgced::puos_editor "PUO Plugin\nFestlegen der PUOs fr alle PGCs in der aktuellen Domain" . false]}]} {
		set ::pgced::puos -1
		set puos [::pgced::puos_editor "PUO Plugin\nFestlegen der PUOs fr alle PGCs in der aktuellen Domain" .]
	}
	if {$puos == -1} return

	set numpgcs 0
	if {$type == "pgc"} {
		if {$vts == 0} {
			set_int ::pgcs(0,0,header) 8 $puos 4
			incr numpgcs
		} else {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set_int ::pgcs($vts,$pgc,header) 8 $puos 4
				incr numpgcs
			} 
		}
	} else {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set_int ::menus($vts,$lu,$pgc,header) 8 $puos 4
				incr numpgcs
			}
		}
	}
	show_msg "Verbotene Benutzeraktionen erfolgreich festgelegt in $numpgcs PGCs."
	::main::dvd_modified true false
}

# Clear VOB PUOs
proc clear_vob_puos {} {
	if {![info exists ::dvddir]} {bell; return}
	set title "PUO Plugin - Entfernen der VOB PUOs          "

	set ::pgced::puos 4194320	;# By default, keep the GoUp and Angle Change PUOs.
	set msg "Filter PUOs fr alle VOBs :\n"
	set msg "${msg}Entfernen Sie die PUOs, die Sie lschen wollen und\n"
	set msg "${msg}legen Sie die PUOs fest, die Sie nicht ndern wollen.\n"
	set msg "${msg}Warnung: Die VOB Dateien werden sofort modifiziert.\n"
	set msg "${msg}Es gibt kein automatisches Backup der VOB Dateien!\n"
	set msg "${msg}Es wird empfohlen, die GoUp und Angle PUOs zu behalten."
	if {[catch {set filterpuos [::pgced::puos_editor $msg . false]}]} {
		set ::pgced::puos -1
		set filterpuos [::pgced::puos_editor $msg . false]
	}
	if {$filterpuos == -1} return

	#if {! $::config(expert_mode) && [tk_messageBox -title $title -icon question -type okcancel \
	#		-default ok -message "Clear all Prohibited User Operations in VOB files?\nThis operation can take a while!\n\nWarning: The VOB files are modified immediately.\nThere is no automatic backup of the VOB files!\n\nContinue?"] == "cancel"} return

	set errors 0
	set last_errors 0
	set ::plugins::iepuos::numnavpacks 0
	set ::plugins::iepuos::numfixed 0

	# compute number of sectors to process for progress bar
	show_msg "Entferne PUOs..."
	set filename [file join $::dvddir "VIDEO_TS.VOB"]
	if {[file exists $filename]} {
		set maxpc [expr [file size $filename] / 2048]
	} else {
		set maxpc 0
	}
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set filename [file join $::dvddir "VTS_[format %02d $vts]_0.VOB"]
		if {[file exists $filename]} {
			incr maxpc [expr [file size $filename] / 2048]
		}
		for {set i 1} {$i < 10} {incr i} {
			set filename [file join $::dvddir "VTS_[format %02d $vts]_$i.VOB"]
			if {[file exists $filename]} {
				incr maxpc [expr [file size $filename] / 2048]
			} else {
				break
			}
		}
	}
	set processed 0

	set filename [file join $::dvddir "VIDEO_TS.VOB"]
	if {[file exists $filename]} {
		set vobsize [expr [file size $filename] / 2048]
		incr processed $vobsize
		show_msg "Entferne PUOs in der VIDEO_TS.VOB..."
		catch {update_progress $processed $maxpc}
		set fh [::menued::openbinfile $filename r+]
		if {$fh != ""} {
			set sector 0
			while {1} {
				set sector [clear_puos_in_nav_pack $filename $fh $sector $filterpuos]
				if {$sector < 0} break
				if {$sector >= $vobsize} break
			}
			close $fh
		}
		if {$sector < -10} {incr errors}
	}
	
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set filename [file join $::dvddir "VTS_[format %02d $vts]_0.VOB"]
		if {[file exists $filename]} {
			set vobsize [expr [file size $filename] / 2048]
			incr processed $vobsize
			show_msg "Entferne PUOs in [file tail $filename]..."
			catch {update_progress $processed $maxpc}
			set fh [::menued::openbinfile $filename r+]
			if {$fh != ""} {
				set sector 0
				while {1} {
					set sector [clear_puos_in_nav_pack $filename $fh $sector $filterpuos]
					if {$sector < 0} break
					if {$sector >= $vobsize} break
				}
				close $fh
			}
			if {$sector < -10} {incr errors}
		}
		if {$errors != $last_errors && [expr $errors % 3] == 0} {
			set last_errors $errors
			if {[tk_messageBox -title $title -icon question -type yesno -default no \
				-message "Bis jetzt $errors gefundene Fehler!\nWollen Sie jetzt abbrechen?"] == "yes"} break
		}

		set vobnum 1
		set sector 0
		set filename [file join $::dvddir "VTS_[format %02d $vts]_$vobnum.VOB"]
		if {[file exists $filename]} {
			set vobsize [expr [file size $filename] / 2048]
			incr processed $vobsize
			show_msg "Entferne PUOs in [file tail $filename]..."
			catch {update_progress $processed $maxpc}
			set fh [::menued::openbinfile $filename r+]
			if {$fh != ""} {
				while {1} {
					set sector [clear_puos_in_nav_pack $filename $fh $sector $filterpuos]
					if {$sector < 0} break
					if {$sector >= $vobsize} {
						close $fh
						set fh ""
						incr vobnum
						set filename [file join $::dvddir "VTS_[format %02d $vts]_$vobnum.VOB"]
						set sector [expr $sector - $vobsize]
						if {[file exists $filename]} {
							set vobsize [expr [file size $filename] / 2048]
							incr processed $vobsize
							show_msg "Entferne PUOs in [file tail $filename]..."
							catch {update_progress $processed $maxpc}
							set fh [::menued::openbinfile $filename r+]
							if {$fh == ""} break
						} else {
							break
						}
					}
				}
				if {$fh != ""} {close $fh}
				if {$sector < -10} {incr errors}
			}
		}
		if {$errors != $last_errors && [expr $errors % 3] == 0} {
			set last_errors $errors
			if {[tk_messageBox -title $title -icon question -type yesno -default no \
				-message "Bis jetzt $errors gefundene Fehler!\nWollen Sie jetzt abbrechen?"] == "yes"} break
		}
	}
	show_msg "VOB PUOs entfernt.  $::plugins::iepuos::numnavpacks Nav Packs gescannt.  $::plugins::iepuos::numfixed Nav Packs korrigiert.  $errors Fehler."
	bell
}
proc clear_puos_in_nav_pack {filename fh sector filterpuos} {
	set navpack [::menued::load_nav_pack $filename $fh $sector]
	if {$navpack == ""} {
		show_msg "*** Nav Pack am Sektor $sector ist ungltig!  Suche nach dem nchsten gltigen Nav Pack..." -1
		set l [search_next_valid_nav_pack $filename $fh $sector]
		set sector  [lindex $l 0]
		set navpack [lindex $l 1]
		if {$navpack == ""} {
			show_msg "*** Kein gltiges Nav Pack mehr gefunden!" -1
			tk_messageBox  -title "Plugin - Verbotene Benutzeraktionen" -icon error \
					-message "Kein gltiges Nav Pack mehr gefunden\nin \"[file tail $filename]\":\nAbbruch in der aktuellen Domain!\nVersuchen Sie, die schlechte(n) VOB Cell (s) zu isolieren oder zu blanken."
			return -11
		}
		show_msg "Ein neues gltiges Nav Pack wurde gefunden am Sektor $sector.  Entferne weiter die PUOs..." -1
	}
	incr ::plugins::iepuos::numnavpacks
	set puos [get_int navpack 0x35 4]
	set newpuos [expr $puos & $filterpuos]
	if {$puos != $newpuos} {
		set_int navpack 0x35 $newpuos 4
		if {! [::menued::save_nav_pack $filename $fh $sector $navpack]} {return -12}
		incr ::plugins::iepuos::numfixed
	}
	set end_of_vobu [get_int navpack 0x040F 4]
	if {$end_of_vobu == 0} {return -1}
	return [expr $sector + $end_of_vobu + 1]
}
proc search_next_valid_nav_pack {filename fh sector} {
	set vobsize [expr [file size $filename] / 2048]
	incr sector
	if {$sector >= $vobsize} {return [list $sector ""]}
	set navpack [::menued::load_nav_pack $filename $fh $sector]
	while {$navpack == ""} {
		incr sector
		if {$sector >= $vobsize} {return [list $sector ""]}
		set navpack [::menued::load_nav_pack $filename $fh $sector]
	}
	return [list $sector $navpack]
}


}	;# end of namespace ::plugins::iepuos
