#############################################################################
# search_and_replace_plugin.tcl Copyright 2005, 2006, 2007, 2008 r0lZ
# PgcEdit plugin to search and replace DVD commands.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "Search and Replace" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.0 or more to use this plugin.
#
# Note: You may customize this file to add predefined search and replace
# actions in the menu.  See example below.
#############################################################################
# Version History:
#
# v 2.6 (December 31, 2007)
# - Bug: some commands were not replaced due to upper/lower case problems.
#
# v 2.5 (November 21, 2007)
# - The directory used to save the presets was wrong again since v2.3.  It is
#   now in Application Data\PgcEdit.  You might have to move your original
#   presets in that folder.  Sorry.
#
# v 2.4 (November 19, 2007)
# - Fixed a bug that prevented the Search and Replace GUI to be opened.
#
# v 2.3 (November 17, 2007)
# - Fixed a possible bug with the Application Data path necessary save the
#   config file (under Windoze only.)
#
# v 2.2 (October 12, 2007)
# - Added the options to limit the action of the Search and Replace to a VTS
#   range, to the menus or title domains, and to the pre, post, cell or button
#   areas only.
#
# v 2.1 (June 20, 2007)
# - When a command is copied from the menu viewer or editor and pasted as the
#   replace string in the Search and Replace GUI, it is enclosed in braces.
#   It was impossible to delete the closing brace before the opening brace.
#
# v 2.0 (June 15, 2007)
# - It is now possible to specify #N fields in the replace string to copy the
#   Nth. byte from the source string at that position.  See help for details.
# - It is also possible to replace a specific byte by the result of a math
#   expression enclosed in parenthesis and containing #N fields.
# - Added Presets menu to open or save the current Search and Replace strings.
# - Added the "Convert SetGPRMMD to Set" hardcoded preset as an useful example
#   of #N substitutions and of the usage of the math expressions.
# - The last search and replace strings used are saved as a default preset,
#   and loaded automatically the first time the GUI is opened.
# - It is now possible to work in PgcEdit when the Search and Replace window
#   is opened.  Handy to copy/paste commands.
# - The contents of the edit boxes are now selected automatically when they
#   receive the focus.
# - The OK button has been replaced by a Replace button.  Now, the Replace
#   window is not closed any more when the "Replace" button is pressed.
# - Added a status bar in the window to display the number of replacements.
#
# - Added the new function "Auto Increment" to increment some parameters of a
#   list of several commands automatically.  Very handy to create, for example,
#   a list of commands to jump to a specific chapter, based on the content of
#   a GPRM.
#
# v 1.1 (September 7, 2005)
# - First public release
#############################################################################


# Add Plugins -> Search & Replace submenu
# create search and replace submenu
if {! [winfo exists .mainmenu.plugins.sandr]} {
	.mainmenu.plugins add cascade -label "Suchen und Ersetzen" -menu .mainmenu.plugins.sandr
	menu .mainmenu.plugins.sandr
	# add Search & Replace with GUI item
	.mainmenu.plugins.sandr add command -label "Suchen und Ersetzen (GUI)" -command \
		{::plugins::sandr::search_and_replace_gui}
	.mainmenu.plugins.sandr add command -label "Automatische Erhhung" -command \
		{::plugins::sandr::auto_increment_gui}
	.mainmenu.plugins.sandr add separator
	.mainmenu.plugins.sandr add command -label "ber..." -command {
			tk_messageBox  -title "Plugin - Suchen und Ersetzen" -icon info \
					-message "Plugin - Suchen und Ersetzen fr PgcEdit 0.6.0 oder hher.\nVersion $::plugins::sandr::version\nAutor: r0lZ\nFreeware!\n\nBemerkung: Sie knnen sich das Men selbst anpassen\nin dem Sie die Tcl-Datei bearbeiten."
		}
}


### Don't edit after this point, unless you know what you are doing!


namespace eval ::plugins {}
namespace eval ::plugins::sandr {

set version "2.6"

# Search and Replace
# This function searches for commands (regular expressions allowed) in the whole DVD
# (including in menu buttons and BOVs) and patches the matching commands with the destination template.
# The destination template must be an hexadecimal command, with dot (.) characters when the
# original digits must be left untouched.
# Usage: ::plugins::sandr::search_and_replace "search pattern" "replace pattern"
# Example:  Replace all JumpTT to Title 1 to JumpTT to Title 2:
# Search for "30 .2 .. .. .. 01 .. .."
# Replace by ".. .. .. .. .. 02 .. .." or by "30 .2 .. .. .. 02 .. .."

proc search_and_replace {src pattern {from 0} {to -1} {include_menus 1} {include_titles 1} \
		{include_pre 1} {include_post 1} {include_cell 1} {include_buttons 1}} {

	if {![info exists ::dvddir]} {bell; return}
	if {$to == -1 || $to > $::pgcs(numvts)} {set to $::pgcs(numvts)}

	set numchanges 0
	status "Ersetze..."
	set errmsg ""
	set numerrors 0

	for {set vts $from} {$vts <= $to} {incr vts} {
		if {$vts == 0} {
		    if {$include_titles} {
				foreach i [lsearch -all -regexp [string toupper $::pgcs(0,0)] $src] {
					if {[subst \$include_[get_cmd_type ::pgcs(0,0) $i]]} {
						set origcmd [lindex [string toupper $::pgcs(0,0)] $i]
						set newcmd [replace_command $origcmd $pattern]
						if {$newcmd == ""} {
							set errmsg "$errmsg\nFP-PGC,\nCommand \"[::disasm::disasm $origcmd]\":\nFehler beim Ersetzen! (*)\n"
							incr numerrors
						}
						set msg [islegal $newcmd cmd 0 0 0 $i]
						if {$msg != ""} {
							set errmsg "$errmsg\nFP-PGC,\nCommand \"[::disasm::disasm $origcmd]\":\n$msg!\n"
							set newcmd ""
							incr numerrors
						}
						if {$origcmd != $newcmd && $newcmd != ""} {
							lset ::pgcs(0,0) $i $newcmd
							incr numchanges
							status "Ersetze... ($numchanges)"
						}
					}
				}
			}
		} else {
		    if {$include_titles} {
				for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
					foreach i [lsearch -all -regexp [string toupper $::pgcs($vts,$pgc)] $src] {
						if {[subst \$include_[get_cmd_type ::pgcs($vts,$pgc) $i]]} {
							set origcmd [lindex [string toupper $::pgcs($vts,$pgc)] $i]
							set newcmd [replace_command $origcmd $pattern]
							if {$newcmd == ""} {
								set errmsg "$errmsg\nVTST $vts, PGC $pgc,\nCommand \"[::disasm::disasm $origcmd]\":\nFehler beim Ersetzen! (*)\n"
								incr numerrors
							}
							set msg [islegal $newcmd cmd $vts 0 $pgc $i]
							if {$msg != ""} {
								set errmsg "$errmsg\nVTST $vts, PGC $pgc,\nCommand \"[::disasm::disasm $origcmd]\":\n$msg!\n"
								set newcmd ""
								incr numerrors
							}
							if {$origcmd != $newcmd && $newcmd != ""} {
								lset ::pgcs($vts,$pgc) $i $newcmd
								incr numchanges
								status "Ersetze... ($numchanges)"
							}
						}
					}
					if {$include_buttons} {
						# BOVs
						set numcells $::pgcs($vts,$pgc,numcells)
						if {$numcells > 0} {
							set cellpostable $::pgcs($vts,$pgc,cellpostable)
							for {set cell 1} {$cell <= $numcells} {incr cell} {
								set idx [expr {($cell-1) * 4}]
								set vobid [get_int cellpostable $idx 2]
								incr idx 3
								set cellid [get_int cellpostable $idx 1]
								if {[info exists ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numbuts)]} {
									set numbuts $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numbuts)
									if {$numbuts > 0} {
										set numgroups $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numgroups)
										if {$numgroups > 0} {
											for {set butnum 1} {$butnum <= $numbuts} {incr butnum} {
												for {set group 1} {$group <= $numgroups} {incr group} {
													set origcmd [string toupper $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,$group,$butnum,command)]
													if {[lsearch -regexp [list $origcmd] $src] > -1} {
														set newcmd [replace_command $origcmd $pattern]
														if {$newcmd == ""} {
															set errmsg "$errmsg\nVTST $vts, PGC $pgc, Cell $cell, BOV $butnum,\nCommand \"[::disasm::disasm $origcmd]\":\nFehler beim Ersetzen! (*)\n"
															incr numerrors
														}
														set msg [islegal $newcmd bov $vts 0 $pgc $butnum]
														if {$msg != ""} {
															set errmsg "$errmsg\nVTST $vts, PGC $pgc, Cell $cell, BOV $butnum,\nCommand \"[::disasm::disasm $origcmd]\":\n$msg!\n"
															set newcmd ""
															incr numerrors
														}
														if {$origcmd != $newcmd && $newcmd != ""} {
															set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,$group,$butnum,command) $newcmd
															set ::buttons::butsinfo(pgc,edited) 1
															set ::buttons::butsinfo(pgc,$vts,edited) 1
															set ::buttons::butsinfo(pgc,$vts,$vobid,edited) 1
															set ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,edited) 1
															incr numchanges
															status "Ersetze... ($numchanges)"
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		if {$include_menus} {
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					foreach i [lsearch -all -regexp [string toupper $::menus($vts,$lu,$pgc)] $src] {
						if {[subst \$include_[get_cmd_type ::menus($vts,$lu,$pgc) $i]]} {
							set origcmd [lindex [string toupper $::menus($vts,$lu,$pgc)] $i]
							set newcmd [replace_command $origcmd $pattern]
							if {$newcmd == ""} {
								set errmsg "$errmsg\nVTSM $vts, LU $lu, PGC $pgc,\nCommand \"[::disasm::disasm $origcmd]\":\nFehler beim Ersetzen! (*)\n"
								incr numerrors
							}
							set msg [islegal $newcmd cmd $vts $lu $pgc $i]
							if {$msg != ""} {
								set errmsg "$errmsg\nVTSM $vts, LU $lu, PGC $pgc,\nCommand \"[::disasm::disasm $origcmd]\":\n$msg!\n"
								set newcmd ""
								incr numerrors
							}
							if {$origcmd != $newcmd && $newcmd != ""} {
								lset ::menus($vts,$lu,$pgc) $i $newcmd
								incr numchanges
								status "Ersetze... ($numchanges)"
							}
						}
					}
					if {$include_buttons} {
						# Menu buttons
						set numcells $::menus($vts,$lu,$pgc,numcells)
						if {$numcells > 0} {
							set cellpostable $::menus($vts,$lu,$pgc,cellpostable)
							for {set cell 1} {$cell <= $numcells} {incr cell} {
								set idx [expr {($cell-1) * 4}]
								set vobid [get_int cellpostable $idx 2]
								incr idx 3
								set cellid [get_int cellpostable $idx 1]
								if {[info exists ::buttons::butsinfo(menu,$vts,$vobid,$cellid,numbuts)]} {
									set numbuts $::buttons::butsinfo(menu,$vts,$vobid,$cellid,numbuts)
									if {$numbuts > 0} {
										set numgroups $::buttons::butsinfo(menu,$vts,$vobid,$cellid,numgroups)
										if {$numgroups > 0} {
											for {set butnum 1} {$butnum <= $numbuts} {incr butnum} {
												for {set group 1} {$group <= $numgroups} {incr group} {
													set origcmd [string toupper $::buttons::butsinfo(menu,$vts,$vobid,$cellid,$group,$butnum,command)]
													if {[lsearch -regexp [list $origcmd] $src] > -1} {
														set newcmd [replace_command $origcmd $pattern]
														if {$newcmd == ""} {
															set errmsg "$errmsg\nVTSM $vts, LU $lu, PGC $pgc, Cell $cell, Button $butnum,\nCommand \"[::disasm::disasm $origcmd]\":\nFehler beim Ersetzen ! (*)\n"
															incr numerrors
														}
														set msg [islegal $newcmd button $vts $lu $pgc $butnum]
														if {$msg != ""} {
															set errmsg "$errmsg\nVTSM $vts, LU $lu, PGC $pgc, Cell $cell, Button $butnum,\nCommand \"[::disasm::disasm $origcmd]\":\n$msg!\n"
															set newcmd ""
															incr numerrors
														}
														if {$origcmd != $newcmd && $newcmd != ""} {
															set ::buttons::butsinfo(menu,$vts,$vobid,$cellid,$group,$butnum,command) $newcmd
															set ::buttons::butsinfo(menu,edited) 1
															set ::buttons::butsinfo(menu,$vts,edited) 1
															set ::buttons::butsinfo(menu,$vts,$vobid,edited) 1
															set ::buttons::butsinfo(menu,$vts,$vobid,$cellid,edited) 1
															incr numchanges
															status "Ersetze... ($numchanges)"
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	if {$numchanges > 0} {
		::main::init_tree_menu_no_move
		::main::dvd_modified true false
	}
	if {$errmsg != ""} {
		bell
		set errmsg "$numerrors Fehler gefunden, whrend der Ausfhrung der Funktion:\nRegular Expression - Suchen: \"$src\"\nVorlage - Ersetzen:           \"$pattern\"\nCommands ersetzt: $numchanges\nCommands nicht ersetzt wegen Fehler: $numerrors\n$errmsg"
		if {[string first "(*)" $errmsg] >= 0} { 
			set errmsg "$errmsg\n\n(*): Der Fehler beim Ersetzen knnte durch einen Ausdruck (Expression) verursacht worden sein,\nder zu einem Wert weniger als 0 oder grer als 255 ausgewertet wurde.\n"
		}
		::main::showtext .sandr_err "Protokoll - Suchen und Ersetzen" $errmsg
	}
	return [list $numchanges $numerrors]
}
proc replace_command {src pat {testmode false}} {
	set b1 [expr 0x[lindex $src 0]]
	set b2 [expr 0x[lindex $src 1]]
	set b3 [expr 0x[lindex $src 2]]
	set b4 [expr 0x[lindex $src 3]]
	set b5 [expr 0x[lindex $src 4]]
	set b6 [expr 0x[lindex $src 5]]
	set b7 [expr 0x[lindex $src 6]]
	set b8 [expr 0x[lindex $src 7]]

	set pat [compact_pat $pat]
	set out {}
	for {set i 0} {$i < 8} {incr i} {
		set f [lindex $pat $i]
		if {[string length $f]==2} {
			if {[string range $f 0 0] == "#"} {
				lappend out "($f)"
			} else {
				if {[string range $f 0 0] == "."} {
					set f [string range [lindex $src $i] 0 0][string range $f 1 1]
				}
				if {[string range $f 1 1] == "."} {
					set f [string range $f 0 0][string range [lindex $src $i] 1 1]
				}
				lappend out "(0x$f)"
			}
		} else {
			lappend out $f
		}
	}

	set out [string map {#1 ${b1} #2 ${b2} #3 ${b3} #4 ${b4} #5 ${b5} #6 ${b6} #7 ${b7} #8 ${b8}} $out]
	set out [subst $out]

	for {set i 0} {$i < 8} {incr i} {
		set f [lindex $out $i]
		if {[string range $f 0 0]=="("} {
			if {! [catch {set ee [expr $f]}]} {
				if {$testmode} {
					lset out $i [format %02X $ee]
				} elseif {$ee < 0 || $ee > 255} {
					return ""
				} else {
					lset out $i [format %02X $ee]
				}
			} else {
				if $testmode {
					lset out $i "<ungltiger Ausdruck>"
				} else {
					return ""
				}
			}
		}
	}
	return $out
}
proc islegal {cmd origin vts lu pgc cmdnum} {
	if {$lu == 0} {
		set type "pgc"
		set cmds $::pgcs($vts,$pgc)
	} else {
		set type "menu"
		set cmds $::menus($vts,$lu,$pgc)
	}
	if {$origin == "cmd"} {
		set origin [get_cmd_type cmds $cmdnum]
	}
	set cmdname [::disasm::cmdname $cmd]
	if {$cmdname == "???"} {
		return "Unbekanntes VM-Command"
	} elseif {! [::ced::valid_command $cmdname $type $vts $origin]} {
		return "$cmdname ist hier ungltig"
	}
	set cmdname [::disasm::cmdname2 $cmd]
	if {$cmdname != ""} {
		if {$cmdname == "???"} {
			return "Unbekanntes VM-Command"
		} elseif {! [::ced::valid_command $cmdname $type $vts $origin]} {
			return "$cmdname ist hier ungltig"
		}
	}
	set disasm [::disasm::disasm $cmd]
	if {[string first "ILLEGAL" $disasm] >= 0} {
		return "Illegaler Parameter in $disasm"
	}
	return ""
}

proc compact_pat {pat} {
	# remove useless spaces
	set pat [string trim $pat]
	while {[string first "  " $pat] > 0} {
		set pat [string map {"  " " "} $pat]
	}
	set out ""
	set inparenthesis 0
	for {set i 0} {$i < [string length $pat]} {incr i} {
		set c [string range $pat $i $i]
		if {$c == "("} {
			incr inparenthesis
		}
		if {$c == ")"} {
			incr inparenthesis -1
			if {$inparenthesis < 0} {return ""}
		}
		if {$c != " " || $inparenthesis == 0} {
			set out $out$c
		}
	}
	if {$inparenthesis != 0} {return ""}
	return $out
}

#### GUI

proc search_and_replace_gui {{src ""} {dst ""}} {
	if {![info exists ::dvddir]} {bell; return}

	if {![info exists ::plugins::sandr::src]} {
		set ::plugins::sandr::src ""
		set ::plugins::sandr::dst ""
	}
	if {$src != "" && $dst != ""} {
		set ::plugins::sandr::src $src
		set ::plugins::sandr::dst $dst
	}
	if {! [info exists ::plugins::sandr::from]} {
		set ::plugins::sandr::from 0
		set ::plugins::sandr::to $::pgcs(numvts)
		set ::plugins::sandr::include_menus 1
		set ::plugins::sandr::include_titles 1
		set ::plugins::sandr::include_pre 1
		set ::plugins::sandr::include_post 1
		set ::plugins::sandr::include_cell 1
		set ::plugins::sandr::include_buttons 1
	}
   	if {$::plugins::sandr::to > $::pgcs(numvts)} {set ::plugins::sandr::to $::pgcs(numvts)}

	set w .sandr
	if {! [winfo exists $w]} {
		menu .search_replace_menu -tearoff false
		.search_replace_menu add cascade -label Voreinstellungen -menu .search_replace_menu.presets
		menu .search_replace_menu.presets -tearoff false
		.search_replace_menu.presets add command -label "ffne Voreinstellung" -command ::plugins::sandr::openpreset
		.search_replace_menu.presets add command -label "Speichere Voreinstellung als..." -command ::plugins::sandr::savepreset
		.search_replace_menu.presets add separator
		.search_replace_menu.presets add command -label "Konvertiere SetGPRMMD zu Set" -command \
			{::plugins::sandr::initpreset "\[45\]3 .. .. .. .. 0. .. .." "(#1+0x1E) .. (((#2&0x0F)==0)?#7:0) #6 #3 #4 (((#2&0x0F)==0)?0:#7) .."}
		.search_replace_menu add command -label "Hilfe" -command ::plugins::sandr::help

		toplevel $w -class Dialog -menu .search_replace_menu
		wm withdraw $w
		wm title $w "Suchen und Ersetzen"
		if {$::config(tk) != "aqua"} {wm transient $w .}
		if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
		wm protocol $w WM_DELETE_WINDOW {set ::plugins::sandr::sandr_done 0}
		bind $w <Escape> "set ::plugins::sandr::sandr_done 0"
		bind $w <Return> "set ::plugins::sandr::sandr_done 1"

		frame $w.f -bd 2 -relief groove
		pack  $w.f -padx 4 -pady 3 -fill x
		label $w.f.srcl -text "Suchen nach Hex-Command: (Reg. Expression)"
		pack  $w.f.srcl -anchor w
		entry $w.f.srce -textvariable ::plugins::sandr::src -width 30 -font $::config(fixed_width_font) \
				-validate all -validatecommand "::plugins::sandr::validate_src_entry %P %V"
		pack  $w.f.srce -padx 4 -fill x
		label $w.f.msg1 -textvariable ::plugins::sandr::msg1 -foreground red -justify left -anchor w -pady 0
		pack  $w.f.msg1 -fill x

		label $w.f.dstl -text "Ersetzen durch: (Benutze . um eine Ziffer auszulassen)"
		pack  $w.f.dstl -anchor w
		entry $w.f.dste -textvariable ::plugins::sandr::dst -width 30 -font $::config(fixed_width_font) \
				-validate all -validatecommand "::plugins::sandr::validate_dst_entry %P %V"
		pack  $w.f.dste -padx 4 -fill x
		label $w.f.msg2 -textvariable ::plugins::sandr::msg2 -foreground red -justify left -anchor w -pady 0
		pack  $w.f.msg2 -fill x

		# range
		frame $w.fr -bd 2 -relief groove
		pack  $w.fr -padx 4 -pady 3 -fill x
		label $w.fr.lab -text "Suchen und Ersetzen einschrnken:"
		pack  $w.fr.lab -anchor w
		frame $w.fr.f1
		pack  $w.fr.f1 -anchor w -pady 2
		label $w.fr.f1.l1 -text "Von VTS"
		pack  $w.fr.f1.l1 -side left
		spinbox $w.fr.f1.s1 -from 0 -to $::pgcs(numvts) -width 2 -textvariable ::plugins::sandr::from \
			-validate all -validatecommand "::plugins::sandr::validate_vts_range from {%d} {%P}"
		pack  $w.fr.f1.s1 -side left
		label $w.fr.f1.l2 -text "bis VTS"
		pack  $w.fr.f1.l2 -side left
		spinbox $w.fr.f1.s2 -from 0 -to $::pgcs(numvts) -width 2 -textvariable ::plugins::sandr::to \
		    -validate all -validatecommand "::plugins::sandr::validate_vts_range to {%d} {%P}"
		pack  $w.fr.f1.s2 -side left

		checkbutton $w.fr.c1 -pady 0 -text "Men Domains einbeziehen" -variable ::plugins::sandr::include_menus
		pack  $w.fr.c1 -anchor w
		checkbutton $w.fr.c2 -pady 0 -text "Titel Domains (und FP-PGC) einbeziehen" -variable ::plugins::sandr::include_titles
		pack  $w.fr.c2 -anchor w
		frame $w.fr.pad
		pack  $w.fr.pad -pady 1
		checkbutton $w.fr.c3 -pady 0 -text "Pre-Commands einbeziehen" -variable ::plugins::sandr::include_pre
		pack  $w.fr.c3 -anchor w
		checkbutton $w.fr.c4 -pady 0 -text "Post-Commands einbeziehen" -variable ::plugins::sandr::include_post
		pack  $w.fr.c4 -anchor w
		checkbutton $w.fr.c5 -pady 0 -text "Cell-Commands einbeziehen" -variable ::plugins::sandr::include_cell
		pack  $w.fr.c5 -anchor w
		checkbutton $w.fr.c6 -pady 0 -text "Button und BOV-Commands einbeziehen" -variable ::plugins::sandr::include_buttons
		pack  $w.fr.c6 -anchor w

		# buttons
		frame  $w.fb
		pack   $w.fb -padx 2 -pady 4 -anchor e
		button $w.fb.ok -text "Ersetzen" -width 10 -command {set ::plugins::sandr::sandr_done 1}
		pack   $w.fb.ok -side left -padx 2
		button $w.fb.cancel -text "Schlieen" -width 10 -command {set ::plugins::sandr::sandr_done 0}
		pack   $w.fb.cancel -side left -padx 2

		entry $w.status -textvariable ::plugins::sandr::status -state readonly -takefocus 0
		pack  $w.status -fill x

		update
		wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
		$w.f.srce configure -width 90
		$w.f.dste configure -width 90
		update
		wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
		set minwidth [string length $::plugins::sandr::src]
		if {[string length $::plugins::sandr::dst] > $minwidth} {
			set minwidth [string length $::plugins::sandr::dst]
		}
		if {$minwidth < 30} {
			set minwidth 30
		}
		$w.f.srce configure -width $minwidth
		$w.f.dste configure -width $minwidth
		update
		set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
		set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
		if {$wx < 10} {set wx 10}
		if {$wy < 22} {set wy 22}
		wm geometry $w +$wx+$wy

		if {$::plugins::sandr::src=="" && $::plugins::sandr::dst==""} {
			openpreset true
		}
	}
	if {$::plugins::sandr::src!="" || $::plugins::sandr::dst!=""} {
		validate_src_entry $::plugins::sandr::src key
		validate_dst_entry $::plugins::sandr::dst key
	}
	wm deiconify $w
	focus $w.f.srce


	set ok false
	while {! $ok} {
		tkwait variable ::plugins::sandr::sandr_done
		set ok true
		if {$::plugins::sandr::sandr_done} {
			validate_src_entry $::plugins::sandr::src key
			validate_dst_entry $::plugins::sandr::dst key
			if {$::plugins::sandr::msg2 != ""} {
				set ok false
				focus $w.f.dste
			}
			if {$::plugins::sandr::msg1 != ""} {
				set ok false
				focus $w.f.srce
			}
			if {! $ok} {
				bell
				continue
			}
			focus .sandr.fb.ok
			set ::plugins::sandr::src [string toupper $::plugins::sandr::src]
			set ::plugins::sandr::dst [string toupper $::plugins::sandr::dst]

			set result [::plugins::sandr::search_and_replace \
					$::plugins::sandr::src \
					$::plugins::sandr::dst \
					$::plugins::sandr::from \
					$::plugins::sandr::to \
					$::plugins::sandr::include_menus \
					$::plugins::sandr::include_titles \
					$::plugins::sandr::include_pre \
					$::plugins::sandr::include_post \
					$::plugins::sandr::include_cell \
					$::plugins::sandr::include_buttons]


			status "[lindex $result 0] Commands ersetzt, [lindex $result 1] Fehler."
			show_msg "Suchen und Ersetzen: \"$::plugins::sandr::src\" zu \"$::plugins::sandr::dst\": [lindex $result 0] Commands ersetzt, [lindex $result 1] Fehler."
			if {[lindex $result 1] == 0} {
				savepreset true
			}
			set ok false
		}
	}
	destroy $w
	destroy .search_replace_menu
	status ""
	unset ::plugins::sandr::pause
}
proc validate_src_entry {str validation_type} {
	if {$validation_type == "focusin"} {
		catch {.sandr.f.srce selection range 0 end}
		catch {.sandr.f.srce icursor end}
		return true
	}
	if {[catch {lsearch -regexp [list "00 00 00 00 00 00 00 00"] $str} errstr]} {
		set ::plugins::sandr::msg1 [string map {"Konnte nicht die Vorlage der Regular Expression kompilieren: " ""} $errstr]!
	} elseif {[string length $str]==0} {
		set ::plugins::sandr::msg1 "Leere Vorlage fr das Suchen!"
	} else {
		set ::plugins::sandr::msg1 ""
	}
	status ""
	return true
}
proc validate_dst_entry {pat validation_type} {
	if {$validation_type == "focusin"} {
		catch {.sandr.f.dste selection range 0 end}
		catch {.sandr.f.dste icursor end}
		return true
	}
	set ::plugins::sandr::msg2 ""

	# remove useless spaces
	if {[string trim $pat] == ""} {
		set ::plugins::sandr::msg2 "Leere Vorlage fr das Ersetzen!"
		return 1
	}
	set pat [compact_pat $pat]
	if {$pat == ""} {
		set ::plugins::sandr::msg2 "Unausgeglichene Klammerung!"
		return 1
	}
	if {[catch {set len [llength $pat]}]} {
		set ::plugins::sandr::msg2 "Unausgeglichene (offene) geschweifte Klammer!"
		return 1
	}
	if {$len > 8} {
		set ::plugins::sandr::msg2 "Zu viele Felder!"
		return 1
	}
	if {$len < 8} {
		set ::plugins::sandr::msg2 "Zu wenige Felder!"
		return 1
	}
	set tr [replace_command {FF FF FF FF FF FF FF FF} $pat true]
	for {set i 0} {$i < 8} {incr i} {
		if {[catch {set f [expr 0x[lindex $tr $i]]}]} {
			set ::plugins::sandr::msg2 "Ungltiges Feld [expr $i+1]!"
			return 1
		}
		if {$f > 510} {
			set ::plugins::sandr::msg2 "Ungltiges Feld [expr $i+1]!"
			return 1
		}
	}
	return 1
}
proc validate_vts_range {source d p} {
	set valid [::ced::value_entry_validate 0 $::pgcs(numvts) $d $p 1]
	if {! $valid} {bell; return false}
	if {$source == "from" && $::plugins::sandr::to < $p} {
	    set ::plugins::sandr::to $p
	}
	if {$source == "to" && $::plugins::sandr::from > $p} {
	    set ::plugins::sandr::from $p
	}
	return true
}


proc status {string} {
	set ::plugins::sandr::status $string
	update idletasks
	after 2 {set ::plugins::sandr::pause 1}
	tkwait variable ::plugins::sandr::pause
}

proc openpreset {{asdefault false}} {
	set defaultdir [getdefaultpresetdir]
	if {$asdefault} {
		set filename [file join $defaultdir "letztes.srp"]
		if {! [file exists $filename]} return
	} else {
		set filename [tk_getOpenFile -initialdir $defaultdir -parent .sandr \
				-title {ffnen der Voreinstellung fr das Suchen und Ersetzen} \
				-filetypes {{{Voreinstellungen - Suchen und Ersetzen} {*.srp}}}]
		if {$filename == ""} return
	}
	if {[catch {source $filename}] && !$asdefault} {
		status "Fehler beim ffnen der Voreinstellung!"
		bell
	}
	validate_src_entry $::plugins::sandr::src key
	validate_dst_entry $::plugins::sandr::dst key
}

proc savepreset {{asdefault false}} {
	if {$::plugins::sandr::msg1 != "" || $::plugins::sandr::msg2 != ""} {
		status "Ungltige Voreinstellung!"
		bell
		return
	}
	set defaultdir [getdefaultpresetdir]
	if {$defaultdir == "" && $asdefault} return
	if {$asdefault} {
		set filename [file join $defaultdir "letztes.srp"]
	} else {
		set filename [tk_getSaveFile -initialdir $defaultdir -parent .sandr \
				-title {Speichern der Voreinstellungen fr das Suchen und Ersetzen} \
				-filetypes {{{Voreinstellungen - Suchen und Ersetzen} {*.srp}}}]
		if {$filename == ""} return
		if {[string toupper [file extension $filename]] != ".SRP"} {
			set filename $filename.srp
		}
	}
	set data "# PgcEdit Search and Replace preset file\n"
	set data "${data}set ::plugins::sandr::src [list $::plugins::sandr::src]\n"
	set data "${data}set ::plugins::sandr::dst [list $::plugins::sandr::dst]\n"
	set rc [saveFile $filename auto $data]
	if {!$asdefault} {
		if {$rc == 0} {
			status "Voreinstellung erfolgreich gespeichert."
		} else {
			status "Fehler beim Speichern der Voreinstellung!"
			bell
		}
	}
	return $rc
}

proc getdefaultpresetdir {} {
	set defaultdir [file join $::config(config_directory) search_replace_presets]
	if {! [file isdirectory $defaultdir]} {
		catch {file mkdir $defaultdir}
	}
	if {! [file isdirectory $defaultdir]} {
		return ""
	}
	return $defaultdir
}

proc initpreset {src dst} {
	set ::plugins::sandr::src $src
	set ::plugins::sandr::dst $dst
	validate_src_entry $::plugins::sandr::src key
	validate_dst_entry $::plugins::sandr::dst key
}

proc help {} {
	set txt "Suchen nach Hex-Commands in der ganzen DVD (einschlielich in den Men-Buttons und BOVs) und die zusammenpassenden"
	set txt "$txt\nCommands durch die Zielvorlage berschreiben."
	set txt "$txt\n\nDas Such-Zeichen muss eine Regular Expression-Vorlage mit einigen bereinstimmenden Hexadecimal-Commands sein (mit Leerfelder zwischen den Bytes)."
	set txt "$txt\n\nDas Ersetzen-Zeichen muss eine Folge von 8 Expressions oder Konstanten sein. Jede Expression wird verwendet, um das entsprechende Byte zu ersetzen."
	set txt "$txt\n\nUm ein Byte durch eine Konstante zu berschreiben, verwenden Sie zwei Ziffern mit hexadecimalen Wert (ohne das vorangehende \"0x\") zwischen 00 und FF."
	set txt "$txt\nWenn ein Punkt im Hexadecimal-Wert anwesend ist, wird die ursprngliche Ziffer an derselben Position im ursprnglichen Byte unberhrt gelassen."
	set txt "$txt\nZum Beispiel: Ersetzen von \"30 .2 .. .. .. 01 .. ..\" durch \".. .. .. .. .. 02 .. ..\" wird alle \"JumpTT to Title 1\" durch \"JumpTT to Title 2\" ersetzen."
	set txt "$txt\n\nZustzlich kann jedes Feld des Ersetzen-Zeichens ein mathematischer Ausdruck in der Klammerung sein."
	set txt "$txt\nIn einem Ersetzen-Zeichen, \"#N\" (worin N eine Nummer zwischen 1 und 8 ist) bezieht sich der Wert auf das \"N\" Byte des Hex-Commands."
	set txt "$txt\nZum Beispiel, mit (#6) an der Position 2 in dem Ersetzen-Zeichen ist gemeint, dass Byte 2 ersetzt werden muss durch das sechste Byte von dem Command."
	set txt "$txt\nMit (#6+0x10) oder (#6+16) ist gemeint, dass das entsprechende Byte des Hex-Commands durch den Wert des 6.Bytes des Commands (plus 16) ersetzt werden muss."
	set txt "$txt\nEs ist sogar mglich, ein Byte bedingungsweise zu ersetzen.  Zum Bespiel, mit (((#2&0x0F)==0)?#7:0) ist gemeint, dass wenn Byte 2 von dem Command"
	set txt "$txt\nkleiner gleich 0 ist, muss das gegenwrtige Byte durch das Byte 7 des Commands ersetzt werden, ansonsten durch 0. (\"00\".)"
	set txt "$txt\nUm einen Wert zu einer konstanten Kalkulation mit mehr als zwei Bytes hinzuzufgen, benutzen Sie zum Beispiel \"((#5+#6+0x10)/256) ((#5+#6+0x10)%256)\"."
	set txt "$txt\n\nHinweise: Wenn ein Feld, einen einziges Ersetzen-Zeichen wie #6 enthlt, ist es nicht notwendig, es zwischen der Klammerung einzuschlieen."
	set txt "$txt\nIn einer Expression sind die gltigen und ntzlichen Operatoren:   +, -, *, / (Integer-Division), % (Restbetrag), & (AND), | (OR) und ^ (XOR)."
	set txt "$txt\nSehen Sie sich auch die fest programmierten und voreingestellten Vorlagen der Funktion \"Konvertiere SetGPRMMD zu Set\" an, um zu sehen,"
	set txt "$txt\nwie man Expressions, Bedingungen und #N Zeichen verwendet."
  set txt "$txt\n\nAufpassen: Diese Funktion kann nicht rckgngig gemacht werden und die Gltigkeitsprfung von dem ersetzten Zeichen ist etwas ungenau!"
	tk_messageBox -title "Plugin - Suchen und Ersetzen: Hilfe fr das Suchen und Ersetzen" -icon info -message $txt
}

####################### Auto Increment ########################
proc auto_increment_gui {} {
	if {![info exists ::dvddir]} {bell; return}

	set idxs [::main::get_selection_indexes]
	set firstcmdn [lindex $idxs 0]
	set lastcmdn  [lindex $idxs 1]
	if {$firstcmdn == $lastcmdn} {
		ihelp true
		return
	}
	set type [lindex $::editor_cmds_from 0]
	set idxs [lindex $::editor_cmds_from 1]
	set cmds [subst \$::${type}s($idxs)]
	set firstcmd [lindex $cmds $firstcmdn]
	if {[string range $firstcmd 0 0] == "*"} {
		ihelp true
		return
	}
	set ::plugins::sandr::isrc [string toupper $firstcmd]
	set ::plugins::sandr::idst [string toupper $firstcmd]

	set w .incrt
	if {! [winfo exists $w]} {
		toplevel $w -class Dialog ;# -menu .search_replace_menu
		wm withdraw $w
		wm title $w "Automatische Erhhung"
		if {$::config(tk) != "aqua"} {wm transient $w .}
		if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
		wm protocol $w WM_DELETE_WINDOW {set ::plugins::sandr::incrt_done 0}
		bind $w <Escape> "set ::plugins::sandr::incrt_done 0"
		bind $w <Return> "set ::plugins::sandr::incrt_done 1"

		frame $w.f -bd 2 -relief groove
		pack  $w.f -padx 4 -pady 3 -fill x

		#label $w.f.srcl -text "Search for hex command: (Reg expr)"
		#pack  $w.f.srcl -anchor w
		#entry $w.f.srce -textvariable ::plugins::sandr::isrc -width 23 -font $::config(fixed_width_font) \
		#		-state readonly
		#pack  $w.f.srce -padx 4 -fill x

		label $w.f.dstl -text "Zeichen zum Erhhen: (Benutzen Sie +1, um ein Byte zu erhhen)"
		pack  $w.f.dstl -anchor w
		entry $w.f.dste -textvariable ::plugins::sandr::idst -width 30 -font $::config(fixed_width_font) \
				-validate all -validatecommand "::plugins::sandr::validate_idst_entry %P %V"
		pack  $w.f.dste -padx 4 -fill x
		label $w.f.msg2 -textvariable ::plugins::sandr::imsg2 -foreground red -justify left -anchor w -pady 0
		pack  $w.f.msg2 -fill x

		# buttons
		frame  $w.fb
		pack   $w.fb -padx 2 -pady 4 -fill x
		button $w.fb.help -text "Hilfe" -command ::plugins::sandr::ihelp
		pack   $w.fb.help -side left -padx 2
		frame  $w.fb.pad -width 2
		pack   $w.fb.pad -side left -fill x -expand 1
		button $w.fb.ok -text "Erhhung" -width 10 -command {set ::plugins::sandr::incrt_done 1}
		pack   $w.fb.ok -side left -padx 2
		button $w.fb.cancel -text "Schlieen" -width 10 -command {set ::plugins::sandr::incrt_done 0}
		pack   $w.fb.cancel -side left -padx 2

		update
		wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
		#$w.f.srce configure -width 90
		$w.f.dste configure -width 90
		update
		wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
		set minwidth [string length $::plugins::sandr::isrc]
		if {[string length $::plugins::sandr::idst] > $minwidth} {
			set minwidth [string length $::plugins::sandr::idst]
		}
		if {$minwidth < 30} {
			set minwidth 30
		}
		#$w.f.srce configure -width $minwidth
		$w.f.dste configure -width $minwidth
		update
		set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
		set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
		if {$wx < 10} {set wx 10}
		if {$wy < 22} {set wy 22}
		wm geometry $w +$wx+$wy
	}
	validate_idst_entry $::plugins::sandr::idst key
	wm deiconify $w
	focus $w.f.dste


	set ok false
	while {! $ok} {
		tkwait variable ::plugins::sandr::incrt_done
		set ok true
		if {$::plugins::sandr::incrt_done} {
			validate_idst_entry $::plugins::sandr::idst key
			if {$::plugins::sandr::imsg2 != ""} {
				set ok false
				focus $w.f.dste
			}
			if {! $ok} {
				bell
				continue
			}
			focus .incrt.fb.ok
			set result [::plugins::sandr::auto_increment $cmds $firstcmdn $lastcmdn [string toupper $::plugins::sandr::isrc] [string toupper $::plugins::sandr::idst]]
			if {$type == "menu"} {
				set ::menus($idxs) $result
			} else {
				set ::pgcs($idxs) $result
			}
			::main::init_tree_menu_no_move
			::main::dvd_modified true false
		}
	}
	destroy $w
	destroy .search_replace_menu
}

proc validate_idst_entry {pat validation_type} {
	if {$validation_type == "focusin"} {
		catch {.incrt.f.dste selection range 0 end}
		catch {.incrt.f.dste icursor end}
		return true
	}
	set ::plugins::sandr::imsg2 ""

	# remove useless spaces
	if {[string trim $pat] == ""} {
		set ::plugins::sandr::imsg2 "Leere Vorlage fr das Ersetzen!"
		return 1
	}
	if {[llength $pat] > 8} {
		set ::plugins::sandr::imsg2 "Zu viele Felder!"
		return 1
	}
	if {[llength $pat] < 8} {
		set ::plugins::sandr::imsg2 "Zu wenige Felder!"
		return 1
	}
	set searchstring ""
	for {set i 0} {$i < 8} {incr i} {
		set f [lindex $pat $i]
		set c [string range $f 0 0]
		if {$c=="+" || $c=="-"} {
			set r [string range $f 1 end]
			if {$r==""} {
				set ::plugins::sandr::imsg2 "Fehlender $c Wert in Feld [expr $i+1]!"
				return 1
			}
			if {![string is integer $r]} {
				set ::plugins::sandr::imsg2 "Ungltiger $c Wert in Feld [expr $i+1]!"
				return 1
			}
			set searchstring "$searchstring .."
		} elseif {[string is xdigit $f]} {
			if {[string length $f] != 2} {
				set ::plugins::sandr::imsg2 "Ungltiges Hexadezimales Feld [expr $i+1]!"
				return 1
			}
			set searchstring "$searchstring $f"
		}
	}
	set ::plugins::sandr::isrc [string trim $searchstring]
	return 1
}

proc auto_increment {cmds firstcmdn lastcmdn src dst} {
	set srccmd [lindex $cmds $firstcmdn]
	set incritems {}
	for {set i 0} {$i < 8} {incr i} {
		set f [lindex $src $i]
		if {$f == ".."} {
			set curval($i) [expr 0x[lindex $srccmd $i]]
			set increment($i) [expr 0[lindex $dst $i]]
			lappend incritems $i
		}
	}
	foreach i [lsearch -all -regexp [string toupper $cmds] [string toupper $src]] {
		if {$i > $firstcmdn && $i <= $lastcmdn} {
			set cmd [lindex $cmds $i]
			foreach j $incritems {
				incr curval($j) $increment($j)
				if {$curval($j) < 0} {
					set curval($j) 0
				}
				if {$curval($j) > 255} {
					set curval($j) 255
				}
				lset cmd $j [format %02X $curval($j)]
			}
			lset cmds $i $cmd
		}
	}
	return $cmds
}

proc ihelp {{error false}} {
	set txt "Diese Funktion ist praktisch, zum Modifizieren einer langen Liste von Commands, wenn ein Wert oder mehrere Werte in jeder"
	set txt "$txt\nReihe erhht werden mssen, wie bei diesen hier:"
	set txt "$txt\n\[20 A5 00 0F 00 01 00 01\]  1  if ( gprm(15) == 1 ) then { LinkPTT Chapter 1 }"
	set txt "$txt\n\[20 A5 00 0F 00 02 00 02\]  2  if ( gprm(15) == 2 ) then { LinkPTT Chapter 2 }"
	set txt "$txt\n\[20 A5 00 0F 00 03 00 03\]  3  if ( gprm(15) == 3 ) then { LinkPTT Chapter 3 }"
	set txt "$txt\n  ..."
	set txt "$txt\nSie knnen jetzt das erste Command einfach erstellen, es mehrere Male kopieren, dann auswhlen und diese Funktion aufrufen."
	set txt "$txt\n\nUm diese Funktion zu verwenden, mssen Sie mehrere gleiche oder identische Commands im aktuellen PGC auswhlen."
	set txt "$txt\nDann knnen Sie das erste Command (hexadezimal) bearbeiten und ein oder mehrere Byte(s) durch die Erhhung ersetzen."
	set txt "$txt\n\nEine Erhhung mu beginnen durch + oder -.  Zum Beispiel, um den GPRM-Wert und die Kapitel-Nummer von dem Command"
	set txt "$txt\n\"if ( gprm(15) == 1 ) then { LinkPTT Chapter 1 }\" zu erhhen, sollten Sie \"20 A5 00 0F 00 +1 00 +1\" benutzen."
	set txt "$txt\nPgcEdit wird dann die Startwerte vom ersten ausgewhlten Command nehmen und die sechsten und letzten Bytes aller hnlichen"
	set txt "$txt\nin der aktuellen Auswahl nachfolgenden enthaltenen Commands erhhen. "
	set txt "$txt\n\nBemerken Sie, dass das erste ausgewhlte Command nicht erhht wird."
	set txt "$txt\n\nDie Bytes, angegeben in dem Ersetzen-Zeichen, die nicht mit + oder - beginnen, werden verwendet um die Commands zu filtern."
	set txt "$txt\nDeshalb werden die Commands, die anders sind als die ersten, automatisch ausgelassen und nicht erhht."
	if {$error} {
		set txt "$txt\n\n\nFehler:  Whlen Sie bitte eine Liste von mehreren hnlichen Commands aus und versuchen Sie es erneut!"
		set icon error
	} else {
		set icon info
	}
	tk_messageBox -title "Plugin - Suchen und Ersetzen: Hilfe zur autom. Erhhung" -icon $icon -message $txt
}

}	;# end of namespace ::plugins::sandr
