#############################################################################
# xiron_plugin.tcl Copyright 2006, 2007 r0lZ
# PgcEdit plugin to fix and modify DVDs created by a XiRON recorder.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "XiRON" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.0 or more to use this plugin.
# It is better to use v 0.6.1 or more.
#############################################################################
# Version History:
#
# v 1.0 (October 6, 2006)
# - First public release
# v 1.1 (May 25, 2007)
# - Little bug fixed: the PGC and GPRM user comments (if any) were not saved
#   when the DVD was saved during Fix VMGM VOB IDs.
#############################################################################



# Add Plugins -> xiron submenu
if {! [winfo exists .mainmenu.plugins.xiron]} {
	.mainmenu.plugins add cascade -label "XiRON" -menu .mainmenu.plugins.xiron
	menu .mainmenu.plugins.xiron

	.mainmenu.plugins.xiron add command -label "Korrigieren aller bekannten Probleme"  \
			-command {::plugins::xiron::fix_all}

	.mainmenu.plugins.xiron add separator
	.mainmenu.plugins.xiron add command -label "1. Korrigiere VOB-IDs der VMGM"  \
			-command {::plugins::xiron::fix_vids}
	.mainmenu.plugins.xiron add command -label "2. Korrigiere Stream-Eigenschaften der Titel"  \
			-command {::plugins::xiron::fix_streams}
	.mainmenu.plugins.xiron add command -label "3. Korrigiere LBA-Adressen der Titel"  \
			-command {::plugins::xiron::fix_lbas}
	.mainmenu.plugins.xiron add command -label "4. Korrigiere LBA-Adressen mit FixVTS"  \
			-command {::plugins::xiron::fix_vts}
	.mainmenu.plugins.xiron add command -label "5. Erstellen der Such-Adressen mit VobBlanker"  \
			-command {::plugins::xiron::build_srptrs}
	.mainmenu.plugins.xiron add command -label "6. Erstellen der Time Maps"  \
			-command {::plugins::xiron::build_tmaps}

	.mainmenu.plugins.xiron add separator
	.mainmenu.plugins.xiron add command -label "Erweitertes Authoring"  \
			-command {::plugins::xiron::enhance}
	.mainmenu.plugins.xiron add command -label "Authoring nochmal"  \
			-command {::plugins::xiron::redo}

	.mainmenu.plugins.xiron add separator
	.mainmenu.plugins.xiron add command -label "ber..." -command {
			tk_messageBox  -title "XiRON Plugin" -icon info \
					-message "XiRON Plugin fr PgcEdit 7.4 oder hher.\nVersion $::plugins::xiron::version\nAutor: r0lZ\nFreeware!"
		}
}

namespace eval ::plugins {}
namespace eval ::plugins::xiron {

############## Modifizierbare Konstanten fr den Benutzer #############
# Pfad zur VobBlanker.exe
set vobblanker_exe "D:\\VobBlanker\\VobBlanker.exe"
# Ordner fr die Ausgabe der Dateien durch den VobBlanker.  
# Die Angabe zwischen den ("") leer lassen, wenn Sie im VobBlanker "Use input folder" benutzen.
set vobblanker_target "E:\\TEST DVD\\TEST 2"
########## Ende der modifizierbaren Konstanten fr den Benutzer ##########

set version "1.2"

# Fix VOB IDs in VIDEO_TS.VOB
proc fix_vids {} {
	set title "XiRON Plugin - Korrigieren der VOB-IDs der VMGM"
	if {![info exists ::dvddir]} {
		show_msg "Keine DVD geladen!"
		bell
		return -1
	}
	if {$::dvd_modified} {
		set txt "Ungespeicherte nderungen gehen verloren!"
		set txt "$txt\nTrotzdem fortfahren?"
		set r [tk_messageBox -message $txt -title $title -icon warning \
				-type okcancel -default cancel]
		if {$r == "cancel"} {
			return -1
		}
	}

	show_msg "Korrigiere VOB-IDs in der VMGM..."
	set filename [file join $::dvddir "VIDEO_TS.VOB"]
	if {![file exists $filename]} {
		show_msg "VIDEO_TS.VOB nicht gefunden!"
		bell
		return -1
	}

	set numsectors [expr [file size $filename]/2048]
	set fh [::menued::openbinfile $filename r+]
	if {$fh == ""} {
		show_msg "VIDEO_TS.VOB kann nicht geffnet werden!"
		bell
		return -1
	}

	set sector 0
	set numfixed 0
	while {1} {
		set navpack [::menued::load_nav_pack $filename $fh $sector]
		if {$navpack == ""} {
			show_msg "Nav Pack am Sektor $sector ist ungltig (in der VIDEO_TS.VOB)!"
			bell
			close $fh
			return -1
		}
		set vid [get_int navpack 0x41F 2]
		if {$vid == 1} {
			show_msg "Nav Pack am Sektor $sector hat schon eine VOB-ID = 1!  (VIDEO_TS.VOB bereits repariert oder falsche DVD?)"
			bell
			close $fh
			return -1
		}
		set new_vid [expr $vid-1]
		set_int navpack 0x41F $new_vid 2
		if {! [::menued::save_nav_pack $filename $fh $sector $navpack]} {
			show_msg "Kann nicht das Nav Pack am LBA $sector in der VIDEO_TS.VOB speichern!"
			bell
			close $fh
			return -1
		}
		incr numfixed

		set end_of_vobu [get_int navpack 0x040F 4]
		if {$end_of_vobu == 0} {
			break
		}
		set sector [expr $sector + $end_of_vobu + 1]
		if {$sector >= $numsectors} {
			break
		}
	}
	close $fh
	::main::ReRead_DVD false true
	show_msg "VMGM VOB-IDs korrigiert, OK.  $numfixed Nav Packs repariert."
	return $numfixed
}

# Fix Titles LBA Pointers
proc fix_lbas {} {
	set title "XiRON Plugin - Korrigieren der LBA-Adressen der Titel"
	if {![info exists ::dvddir]} {
		show_msg "Keine DVD geladen!"
		bell
		return -1
	}
	if {$::dvd_modified} {
		set txt "Ungespeicherte nderungen gehen verloren!"
		set txt "$txt\nTrotzdem fortfahren?"
		set r [tk_messageBox -message $txt -title $title -icon warning \
				-type okcancel -default cancel]
		if {$r == "cancel"} {
			return -1
		}
	}
	set error [get_vobu_lbas 1 lbas]
	if {$error != 0} {
		return -1
	}

	show_msg "Repariere die IFO der PGCs..."
	for {set pgc 1} {$pgc <= $::pgcs(1,numpgcs)} {incr pgc} {
		set numcells $::pgcs(1,$pgc,numcells)
		set postable $::pgcs(1,$pgc,cellpostable)
		for {set c 1} {$c <= $numcells} {incr c} {
			set cellposidx [expr {($c-1)*4}]
			set vid [get_int postable $cellposidx 2]
			set cid [get_int postable [expr {$cellposidx+3}] 1]
			# Replaces the dummy, inexistant cell 1/255 by cell 1/1
			if {$vid==1 && $cid==255 && ![info exists lbas($vid,$cid,entry)]} {
				set cid 1
				set_int postable [expr {$cellposidx+3}] $cid 1
				set ::pgcs(1,$pgc,cellpostable) $postable
			}

			set cellpbidx [expr {($c-1)*24}]
			if {[info exists lbas($vid,$cid,entry)]} {
				set_int ::pgcs(1,$pgc,cellpbtable) [expr $cellpbidx+8]  $lbas($vid,$cid,entry) 4
				set_int ::pgcs(1,$pgc,cellpbtable) [expr $cellpbidx+16] $lbas($vid,$cid,laststart) 4
				set_int ::pgcs(1,$pgc,cellpbtable) [expr $cellpbidx+20] $lbas($vid,$cid,lastend) 4
			}
		}
	}

	show_msg "Repariere VTS_C_ADT..."
	set table $::raw_sectors(1,VTS_C_ADT)
	set len [string length $table]
	set numfixed 0
	for {set idx 8} {$idx < $len} {incr idx 12} {
		set vid [get_int table $idx 2]
		set cid [get_int table [expr {$idx+2}] 1]
		if {[info exists lbas($vid,$cid,entry)]} {
			if {[get_int table [expr {$idx+4}] 4] != $lbas($vid,$cid,entry) || \
				[get_int table [expr {$idx+8}] 4] != $lbas($vid,$cid,lastend)} {
					set_int table [expr {$idx+4}] $lbas($vid,$cid,entry) 4
					set_int table [expr {$idx+8}] $lbas($vid,$cid,lastend) 4
					incr numfixed
			}
		}
	}
	set ::raw_sectors(1,VTS_C_ADT) $table
	::main::dvd_modified true false
	show_msg "VTS1 fertig.  Repariere schlechte VOBU-Adressen von $numfixed Cells."
}

proc build_tmaps {} {
	if {![info exists ::dvddir]} {
		show_msg "Keine DVD geladen!"
		bell
		return -1
	}
	# rebuild the time maps
	show_msg "Erstellen der Time Maps..."
	::tmap::rebuild_tmap_from_vobs 1
	::tmap::dont_need_new_tmap 1

	# Remove Time Search PUO in Title Play Map table
	show_msg "Entfernen der PUOs (UOP0 und UOP1) in Title Play Map Tabelle."
	set table $::raw_sectors(0,VMG_TT_SRPT)
	set numtitles [get_int table 0 2]
	for {set t 0} {$t < $numtitles} {incr t} {
		set idx [expr 8+($t*12)]
		set_int table $idx [expr [get_int table $idx 1] & 0xFC] 1
	}
	set ::raw_sectors(0,VMG_TT_SRPT) $table
	::main::dvd_modified true false
}

proc fix_streams {} {
	set default_lang_code "fr"	;# User modifiable constant

	if {![info exists ::dvddir]} {
		show_msg "Keine DVD geladen!"
		bell
		return -1
	}
	show_msg "Korrigiere Streams."
	set mat_table $::raw_sectors(1,VTSI_MAT)
	set numaudios  [get_int mat_table [expr {0X200+2}] 2]
	set numsubpics [get_int mat_table [expr {0X200+0X54}] 2]
	if {$numaudios != 1 || $numsubpics > 1} {
		show_msg "Die Anzahl von den Audio und/oder Subpic-Streams ist nicht 1!  Falsche DVD?"
		bell
		return -1
	}

	# Source picture already letterboxed?
	set byte1 [get_int mat_table 0x201 1]
	if {[expr $byte1 & 4] > 0} {
		set default "yes"
	} else {
		set default "no"
	}
	set txt "Ist das Quellbild bereits Letterboxed?"
	set txt "$txt\n(Antwort \"Ja\" wenn das Bild horizontale schwarze Balken hat.)"
	set txt "$txt\nAktueller Wert: $default"
	set r [tk_messageBox -message $txt -title "XiRON Plugin - Korrigieren der Streams" \
			-icon question -type yesnocancel -default $default]
	if {$r == "cancel"} {
		show_msg "Video ist bereits Letterboxed, nderung bersprungen!"
	} else {
		show_msg "Einstellen des Videos mit dem Letterboxed Flag zu: $r"
		if {$r == "yes"} {
			set byte1 [expr $byte1 | 4]
		} else {
			set byte1 [expr $byte1 & 0xFB]
		}
		set_int mat_table 0x201 $byte1 1
	}

	# Get AUDIO lang code
	set aattr [string range $mat_table [expr 0X200+4] [expr 0X200+11]]
	set lang_code [lindex [::streams::audio_attributes $aattr 0] 0]
	set orig_lang_code $lang_code
	if {$lang_code == "--"} {set lang_code $default_lang_code}
	show_msg "Ursprnglicher Audiosprachcode: $orig_lang_code"

	set txt "Eingeben des Audiosprachcodes mit"
	set txt "$txt\n2 Buchstaben in Kleinschrift (\"en\", \"de\" ...)"
	set txt "$txt\noder Leerzeichen um den Code zu entfernen."
	set txt "$txt\nAktueller Wert: $orig_lang_code"
	set r [::stringreq::stringreq -window .xiron_stream -parent . \
			-title "XiRON Plugin - Korrigieren der Streams" -text $txt \
			-def $lang_code -callback ::streams::validate_langcode \
			-width 2 -geometry "c" -font $::config(fixed_width_font) \
			-cancelreturn "cancel"]
	if {$r == "cancel"} {
		show_msg "nderung des Audiosprachcodes bersprungen!"
	} else {
		if {$r == ""} {set r \000\000}
		if {$r == $orig_lang_code || [string length $r] != 2} {
			show_msg "Audiosprachcode nicht gendert."
		} else {
			show_msg "Audiosprachcode gendert zu: $r"
			set mat_table [string range $mat_table 0 517]$r[string range $mat_table 520 end]
			set byte0 [get_int mat_table [expr 0X200+4] 1]
			if {$r == "\000\000"} {
				set_int mat_table [expr 0x200+4] [expr $byte0 & 0xF3] 1
			} else {
				set_int mat_table [expr 0x200+4] [expr $byte0 | 0x04] 1
			}
		}
	}

	# Remove subpic definition from mat table
	show_msg "Entferne Subpic-Stream..."
	set_int mat_table [expr 0X200+0x54] 0 2
	set mat_table [string range $mat_table 0 [expr 0X200+0x55]]\000\000\000\000\000\000[string range $mat_table [expr 0X200+0x5C] end]
	# and remove stream assignments in PGCs
	for {set pgc 1} {$pgc <= $::pgcs(1,numpgcs)} {incr pgc} {
		set ::pgcs(1,$pgc,header) \
			[string range $::pgcs(1,$pgc,header) 0 27][string repeat \000 128][string range $::pgcs(1,$pgc,header) 156 end]
	}

	# Fix NTSC/PAL standard in VTSM domain
	if {[expr [get_int mat_table 0x200 1] & 0x30] == 16} {
		show_msg "Erzwinge den Videostandard PAL fr VTSM 1"
		set mode [expr [get_int mat_table 0x100 1] | 0x10]
		set_int mat_table 0x100 $mode 1
	}


	set ::raw_sectors(1,VTSI_MAT) $mat_table
	::main::dvd_modified true false
	show_msg "Fertig.  VTSI_MAT Tabelle und PGCs korrigiert."
	return 0
}

proc fix_vts {{start false}} {
	if {![info exists ::dvddir]} {
		show_msg "Keine DVD geladen!"
		bell
		return -1
	}
	::main::display_pgc pgc 1 1
	::toolscfg::fixvts $start true
}

proc build_srptrs {} {
	if {![info exists ::dvddir]} {
		show_msg "Keine DVD geladen!"
		bell
		return -1
	}
	show_msg "Starten des VobBlanker..."
	# launch VobBlanker
	# Retrieve VobBlanker.exe filename
	if {! [file exists $::plugins::xiron::vobblanker_exe]} {
		show_msg "Pfad zur VobBlanker.exe nicht konfiguriert.  Bearbeiten Sie bitte die xiron_plugin.tcl manuell!"
		bell
		return false
	}
	# Generate INI file for VobBlanker
	set out "\[General\]\n"
	set out "${out}InputPath=[file nativename $::dvddir]\n"
	if {$::plugins::xiron::vobblanker_target == ""} {
		set target [file nativename $::dvddir]
		set check_folder 1
	} else {
		set target $::plugins::xiron::vobblanker_target
		set check_folder 0
	}
	set out "${out}OutputPath=$target\n"
	set out "${out}CheckMenu=1\n"
	set out "${out}CheckTitle=1\n"
	set out "${out}CheckIfoPuops=0\n"
	set out "${out}CheckVobPuops=0\n"
	# may be needed when a V/CID has been changed in the VOBs
	set out "${out}CheckFolder=$check_folder\n"

	# Options moved in subsections since VB 2.1,
	# but kept here also for compatibility with older versions
	#
	# In PGC options section:
	set out "${out}CheckPostPre=0\n"
	set out "${out}CheckPostPreForced=0\n"
	set out "${out}CheckNextMenu=0\n"
	set out "${out}CheckNextTitle=0\n"
	set out "${out}CheckPrevMenu=0\n"
	set out "${out}CheckPrevTitle=0\n"
	set out "${out}CheckAttrTitle=0\n"
	# In FixVOBPointers section:
	set out "${out}CheckFixPointers=1\n"

	# New section in VB 2.1
	set out "${out}\[PGC options\]\n"
	set out "${out}CheckPostPre=0\n"
	set out "${out}CheckPostPreForced=0\n"
	set out "${out}CheckNextMenu=0\n"
	set out "${out}CheckNextTitle=0\n"
	set out "${out}CheckPrevMenu=0\n"
	set out "${out}CheckPrevTitle=0\n"
	set out "${out}CheckAttrTitle=0\n"

	# New section in VB 2.1
	set out "${out}\[FixVOBPointers\]\n"
	set out "${out}CheckFixTitlePointers=1\n"
	set out "${out}CheckFixMenuPointers=1\n"

	set out "${out}\[ScanVOBUs\]\n"
	set out "${out}CheckAutoScan=1\n"

	set out "${out}\[VobLayout\]\n"
	set out "${out}CheckVobFill=0\n"
	set out "${out}CheckVobOrder=1\n"

	set out "${out}CheckVobMUnref=0\n"
	set out "${out}CheckVobLUnref=0\n"
	set out "${out}CheckVobRenumber=1\n"
	set out "${out}CheckClockDisc=1\n"

	set out "${out}\[VIDEO_TS\]\n"
	set out "${out}ActionVTS=PROCESS\n"

	set out "${out}\[VTS01\]\n"
	set out "${out}ActionVTS=PROCESS\n"
	#set out "${out}\[Backup\]\n"
	#set out "${out}CheckBckVts=1\n"

	# Write INI file
	set inifilename [file join $::config(tempdir) "VobBlanker_xiron.INI"]
	if {[saveFile $inifilename auto $out] != 0} {
		return false
	}
	# launch VobBlanker
	wm iconify .
	set rc [catch {exec -- $::plugins::xiron::vobblanker_exe $inifilename} errstr]
	wm deiconify .
	catch {file delete $inifilename}
  if {$rc} {
		tk_messageBox -title "VobBlanker - Fehler" -message $errstr -icon error
		return false
	} else {
		return true
	}
}

proc fix_all {} {
	if {![info exists ::dvddir]} {
		show_msg "Keine DVD geladen!"
		bell
		return -1
	}
	fix_vids

	fix_streams
	show_msg "Speichere DVD..."
	::main::Write_DVD false

	fix_lbas
	show_msg "Speichere DVD..."
	::main::Write_DVD false

	show_msg "Starte FixVTS mit der Option -p"
	fix_vts true

	build_tmaps
	show_msg "Speichere DVD..."
	::main::Write_DVD false

	# finish with VobBlanker process
	build_srptrs
}

proc get_vobu_lbas {vts lbas_upvar} {
	upvar lbas $lbas_upvar
	if {![info exists ::dvddir]} {bell; return -1}

	set error 0

	# compute number of sectors to process for progress bar
	show_msg "Lesen der VOBU LBAs von den Cells..."
	set numsectors 0
	for {set i 1} {$i < 10} {incr i} {
		set filename [file join $::dvddir "VTS_[format %02d $vts]_$i.VOB"]
		if {[file exists $filename]} {
			incr numsectors [expr [file size $filename] / 2048]
		} else {
			break
		}
	}

	set vobnum 1
	set lba 0
	set sector 0
	set lastlba -1
	set vid -1
	set cid -1
	set filename [file join $::dvddir "VTS_[format %02d $vts]_$vobnum.VOB"]
	if {![file exists $filename]} {
		show_msg "$filename nicht gefunden!"
		bell
		return -1
	}
	#show_msg "Opening file [file tail $filename]"
	set vobsize [expr [file size $filename] / 2048]
	show_msg "Lesen der VOBU LBAs von den Cells in [file tail $filename]..." $lba $numsectors
	set fh [::menued::openbinfile $filename r+]
	if {$fh != ""} {
		while {1} {
			set navpack [::menued::load_nav_pack $filename $fh $sector]
			if {$navpack == ""} {
				show_msg "Nav Pack am Sektor $lba ist ungltig (in [file tail $filename] am Sektor $sector)!" $lba $numsectors

				set l [search_next_valid_nav_pack $filename $fh $sector]
				set lba [expr $lba + ([lindex $l 0]-$sector)]
				set sector  [lindex $l 0]
				set navpack [lindex $l 1]
				if {$navpack == ""} {
					if {$sector < $vobsize} {
						set error -1
					}
					break
				}
				show_msg "Ein neues gltiges Nav Pack wurde gefunden am Sektor $sector." $lba $numsectors
			}
			set vid [get_int navpack 0x41F 2]
			set cid [get_int navpack 0x422 1]
			#puts "$vid/$cid at LBA $lba (in [file tail $filename] at $sector)"
			if {![info exists lbas($vid,$cid,entry)]} {
				show_msg "VOB/CELL-ID $vid/$cid gefunden am LBA $lba (in [file tail $filename] am Sektor $sector)" $lba $numsectors
				set lbas($vid,$cid,entry) $lba
				if {$lastlba != -1} {
					set lbas($lastvid,$lastcid,laststart) $lastlba
					set lbas($lastvid,$lastcid,lastend) [expr $lba-1]
				}
				set lastvid $vid
				set lastcid $cid
			}

			set end_of_vobu [get_int navpack 0x040F 4]
			# XiRION bug: wrong VOBU end address
			if {$cid == 255} {set end_of_vobu 0}
			set sector [expr $sector + $end_of_vobu + 1]
			set lastlba $lba
			set lba    [expr $lba    + $end_of_vobu + 1]
			if {$lba >= $numsectors || $sector < 0} {
				break
			}
			if {$sector >= $vobsize} {
				close $fh
				set fh ""
				incr vobnum
				set filename [file join $::dvddir "VTS_[format %02d $vts]_$vobnum.VOB"]
				set sector [expr $sector - $vobsize]
				if {[file exists $filename]} {
					#show_msg "Opening file [file tail $filename]"
					set vobsize [expr [file size $filename] / 2048]
					show_msg "Lesen der VOBU LBAs von den Cells in [file tail $filename]..." $lba $numsectors
					set fh [::menued::openbinfile $filename r+]
					if {$fh == ""} break
				} else {
					break
				}
			}
		}
		if {$fh != ""} {close $fh}
		if {$lastlba != -1} {
			set lbas($lastvid,$lastcid,laststart) $lastlba
			set lbas($lastvid,$lastcid,lastend) [expr $lba-1]
		}
		show_msg "Fertig mit dem Korrigieren der VOBU LBAs von VTST $vts."
	} else {
		set error -1
	}
	return $error
}

proc search_next_valid_nav_pack {filename fh sector} {
	set vobsize [expr [file size $filename] / 2048]
	incr sector
	if {$sector >= $vobsize} {return [list $sector ""]}
	set navpack [::menued::load_nav_pack $filename $fh $sector]
	while {$navpack == ""} {
		incr sector
		if {$sector >= $vobsize} {return [list $sector ""]}
		set navpack [::menued::load_nav_pack $filename $fh $sector]
	}
	return [list $sector $navpack]
}


# Enhance authoring
proc enhance {} {
	if {![info exists ::dvddir]} {bell; return}
	# how many real Titles?
	set numtitles [number_of_real_titles]
	if {$numtitles == -1} return

	# return to menu after playing a regular title
	for {set p 1} {$p <= $numtitles} {incr p} {
		set_type_cmds post ::pgcs(1,$p) {{30 08 00 01 01 C0 00 00}}
	}
	# select Title 1 button when last title has finished playing
	foreach p [list $numtitles [expr $numtitles*2]] {
		set ::pgcs(1,$p) {{* pre} {71 00 00 04 00 01 00 00} {30 08 00 01 01 C0 00 00} {* post} {30 08 00 01 01 C0 00 00} {* cell}}
	}
	# ... and modify the "DVD" button to jump to the first title in loop mode
	if {[info exists ::buttons::butsinfo(menu,0,1,1,1,1,command)] && \
			$::buttons::butsinfo(menu,0,1,1,1,1,command) == "20 01 00 00 00 00 04 05"} {
		set ::buttons::butsinfo(menu,0,1,1,1,1,command) "30 02 00 00 00 [format %02X [expr $numtitles+1]] 00 00"
		set ::buttons::butsinfo(menu,edited) 1
		set ::buttons::butsinfo(menu,0,edited) 1
		set ::buttons::butsinfo(menu,0,1,edited) 1
		set ::buttons::butsinfo(menu,0,1,1,edited) 1
	}

	# Move Title menu from VMGM PGC 1 to 2,
	# so that pressing TITLE calls now the next title.
	# ROOT still calls the menu.
	::utils::set_menu_type 0 1 2 2

	# set NextPGCN link to itself
	for {set p 1} {$p <= $::pgcs(1,numpgcs)} {incr p} {
		set_int ::pgcs(1,$p,header) 156 $p 2
	}

	# Change color of button highlights
	set_int ::menus(0,1,1,header) 0xB1 0xE0748600 4

	# set provider ID
	set newid "XIRION_PGCEDIT"
	set newid "$newid[string repeat \000 [expr {32-[string length $newid]}]]"
	set ::raw_sectors(0,VMGM_MAT) \
			[string replace $::raw_sectors(0,VMGM_MAT) 0X40 0X5F $newid]

	::main::dvd_modified true false
	::main::display_pgc pgc 0 0
	show_msg "Erweitertes Authoring erfolgreich.  Speichern Sie bitte die DVD!"

	# input DVD name
	::utils::set_General_Name_GUI
}

# Redo authoring
proc redo {} {
	if {![info exists ::dvddir]} {bell; return}
	# how many real Titles?
	set numtitles [number_of_real_titles]
	if {$numtitles == -1} return

	#set ::warning_delete_last_pgc true
	for {set p $::pgcs(1,numpgcs)} {$p >= $numtitles} {incr p -1} {
		::main::display_pgc pgc 1 1
		::utils::delete_last_pgc true
	}
	#set ::warning_delete_last_pgc false
	::utils::delete_all_DVD_commands true

	# we don't need the dummy title
	incr numtitles -1

	# set NextPGCN link to itself
	for {set p 1} {$p <= $::pgcs(1,numpgcs)} {incr p} {
		set_int ::pgcs(1,$p,header) 156 $p 2
	}

	##### Build commands #####
	### FP-PGC
	set ::pgcs(0,0) {
		{* pre}
		{71 00 00 00 FF FF 00 00}
		{30 02 00 00 00 01 00 00}
		{* post}
		{* cell}
	}

	### VMGM 1
	set tmp {
		{00 B1 00 00 FF FF 00 04}
		{71 00 00 00 00 00 00 00}
		{20 06 00 00 00 00 04 01}
		{00 00 00 00 00 00 00 00}
		{71 00 00 01 00 01 00 00}
		{00 00 00 00 00 00 00 00}
		{61 00 00 0F 00 84 00 00}
		{74 00 00 0F 00 01 00 00}
		{77 00 00 0F 00 03 00 00}
		{73 00 00 0F 00 01 00 00}
		{75 00 00 0F 04 00 00 00}
		{46 00 00 00 00 0F 00 00}
		{00 00 00 00 00 00 00 00}
		{61 00 00 0F 00 84 00 00}
		{74 00 00 0F 00 01 00 00}
		{76 00 00 0F 00 03 00 00}
		{73 00 00 0F 00 02 00 00}
	}
	for {set i 2} {$i <= [expr ($numtitles/3)+2]} {incr i} {
		lappend tmp "20 A6 00 0F 00 [format %02X $i] 00 [format %02X $i]"
	}
	lappend tmp {00 00 00 00 00 00 00 00} {56 00 00 00 04 00 00 00}
	set_type_cmds pre  ::menus(0,1,1) $tmp
	set_type_cmds post ::menus(0,1,1) {
		{71 00 00 01 00 00 00 00}
		{20 A1 00 02 00 01 00 10}
		{30 02 00 00 00 01 00 00}
	}
	# Enable RESUME with the GoUp key
	set_int ::menus(0,1,1,header) 0xA0 0xFFFF 2
	# Change color of button highlights
	set_int ::menus(0,1,1,header) 0xB1 0xE0748600 4
	# Modify the "DVD" button to enable loop mode: LinkTailPGC
	if {[info exists ::buttons::butsinfo(menu,0,1,1,1,1,command)]} {
		set ::buttons::butsinfo(menu,0,1,1,1,1,command) {20 01 00 00 00 00 00 0D}
		set ::buttons::butsinfo(menu,edited) 1
		set ::buttons::butsinfo(menu,0,edited) 1
		set ::buttons::butsinfo(menu,0,1,edited) 1
		set ::buttons::butsinfo(menu,0,1,1,edited) 1
	}
	# Modify the last (dummy) Title button command to go to main DVD menu page
	set but [expr ($numtitles%3)+1]
	set vid [expr ($numtitles/3)+2]
	if {[info exists ::buttons::butsinfo(menu,0,$vid,1,1,$but,command)]} {
		set ::buttons::butsinfo(menu,0,$vid,1,1,$but,command) {20 06 00 00 00 00 04 01}
		set ::buttons::butsinfo(menu,0,$vid,edited) 1
		set ::buttons::butsinfo(menu,0,$vid,1,edited) 1
	}

	### VMGM 2 (Title menu)

	# Move Title menu from VMGM PGC 1 to 2,
	# so that pressing TITLE calls now the next title.
	# ROOT still calls the menu.
	::utils::set_menu_type 0 1 2 2

	set tmp {
		{71 00 00 01 00 00 00 00}
		{61 00 00 0F 00 84 00 00}
		{73 00 00 0F 00 01 00 00}
	}
	lappend tmp	"00 E1 00 0F 00 [format %02X $numtitles] 00 06" \
			{20 04 00 00 00 00 00 01} \
			{00 00 00 00 00 00 00 00} \
			{71 00 00 0E 00 00 00 00}
	for {set i 1} {$i <= $numtitles} {incr i} {
		lappend tmp {73 00 00 0E 00 01 00 00} "30 22 00 00 00 [format %02X $i] 0E 0F"
	}
	set_type_cmds pre ::menus(0,1,2) $tmp

	### VMGM 3
	set_type_cmds pre ::menus(0,1,3) {
		{20 A4 00 01 00 00 00 02}
		{20 04 00 00 00 00 00 01}
	}

	### VTST 1 (Root menu)
	set_type_cmds pre ::menus(1,1,1) {
		{71 00 00 01 00 01 00 00}
		{30 06 00 01 00 C0 00 00}
	}

	### Titles
	for {set p 1} {$p <= $numtitles} {incr p} {
		set_type_cmds pre  ::pgcs(1,$p) {
			{71 00 00 02 00 01 00 00}
		}
		set_type_cmds post ::pgcs(1,$p) {
			{71 00 00 02 00 00 00 00}
			{30 08 00 03 01 C0 00 00}
		}
	}
	# Overwrite Title 1 pre-commands
	set_type_cmds pre ::pgcs(1,1) {
		{00 B1 00 00 FF FF 00 03}
		{30 08 00 01 01 C0 00 00}
		{00 00 00 00 00 00 00 00}
		{71 00 00 02 00 01 00 00}
	}

	##### Misc #####
	# set provider ID
	set newid "XIRION_PGCEDIT"
	set newid "$newid[string repeat \000 [expr {32-[string length $newid]}]]"
	set ::raw_sectors(0,VMGM_MAT) \
			[string replace $::raw_sectors(0,VMGM_MAT) 0X40 0X5F $newid]

	::main::dvd_modified true false
	::main::display_pgc pgc 0 0
	show_msg "Nochmaliges Authoring erfolgreich.  Speichern Sie bitte die DVD!"

	# input DVD name
	::utils::set_General_Name_GUI
}

proc number_of_real_titles {} {
	if {! [info exists ::menus(0,1,2)]} {
		show_msg "Fehlende VMGM oder falsches Format!  Falsche DVD?"
		bell
		return -1
	}
	set numtitles 0
	set cmds $::menus(0,1,2)
	set cmds [lsearch -all -inline -regexp $cmds "30 22 00 00 00 .. 04 01"]
	if {[llength $cmds] == 0} {
		show_msg "VMGM hat kein korrektes Format!  Falsche DVD?"
		bell
		return -1
	}
	set numtitles [expr 0x[lindex [lindex $cmds end] 5]]
	if {$numtitles != [expr $::pgcs(1,numpgcs)/2]} {
		show_msg "Kann nicht die Anzahl von den wirklichen Titeln bestimmen! Falsche DVD?"
		bell
		return -1
	}
	return $numtitles
}

}	;# end of namespace ::plugins::xiron
